# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
import unittest

from towhee.engine.task_executor import TaskExecutor
from towhee.engine.task_scheduler import FIFOTaskScheduler
from towhee.tests.mock_pipelines import EmulatedPipeline


@unittest.skip('Scheduler changed')
class TestFIFOTaskScheduler(unittest.TestCase):
    """Basic test case for `FIFOTaskScheduler`.
    """

    def setUp(self):
        cache_path = Path(__file__).parent.parent.resolve()
        self._task_execs = []
        for _ in range(4):
            self._task_execs.append(TaskExecutor('', cache_path=cache_path))
        self._pipeline = EmulatedPipeline()
        self._task_sched = FIFOTaskScheduler(self._task_execs)

    def tearDown(self):
        for task_exec in self._task_execs:
            task_exec.stop()

        # Ensure that exactly `n_runs` tasks have been completed after joining.
        self.assertEqual(self._runs_count, self._n_runs)

    def test_scheduler(self, n_runs=1024):

        # Add graph contexts to scheduler.
        for graph_ctx in self._pipeline.graph_contexts:
            self._task_sched.register(graph_ctx)

        # Add callback function upon completion.
        self._n_runs = n_runs
        self._runs_count = 0

        def _add_task_finish_callback(task):
            self._runs_count += 1  # No need for threading.Lock due to GIL.
            self.assertEqual(task.outputs.sum, task.inputs['num'])
        self._pipeline.add_task_finish_handler(_add_task_finish_callback)

        # Spin up executors.
        for task_exec in self._task_execs:
            task_exec.start()

        # Add callback function upon completion.
        for _ in range(n_runs):
            self._task_sched.schedule_step()


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import timeit
import logging
from typing import Any, Callable, Dict, List, NamedTuple, Tuple

from towhee.operator import Operator


class Task:
    """Tasks represent containers for which input, output, and operator data is stored.
    Tasks themselves have no functioning logic, and merely serve as a container for the
    `TaskExecutor`.

    Args:
        op_name:
            Instead of each task executing its own operator, the Task maintains a
            operator name which can be used by executors to lookup the proper `Operator`
            to execute.
        hub_op_id:
            Each operator in the operator hub is tagged with a string that describes its
            functionality.
        op_args:
            Input initialization arguments to the operator. Operators with the same tag
            but with different initialization arguments are considered separate
            operators within the same graph.
        inputs:
            A dictionary of keyward arguments which serve as inputs to the operator
            call.
        task_idx:
            A new task will be constructed for each operator call. The tasks
            are indexed individually for each operation performed, starting from 0.
    """

    def __init__(self, op_name: str, hub_op_id: str, op_args: Dict[str, Any],
                 inputs: Dict[str, Any], task_idx: int):
        self._op_name = op_name
        self._hub_op_id = hub_op_id
        self._op_args = op_args
        self._inputs = inputs
        self._task_idx = task_idx

        self._outputs = None
        self._runtime = -1

        self._on_ready_handlers = []
        self._on_start_handlers = []
        self._on_finish_handlers = []

    @property
    def op_name(self) -> str:
        return self._op_name

    @property
    def hub_op_id(self) -> str:
        return self._hub_op_id

    @property
    def op_args(self) -> Dict[str, Any]:
        return self._op_args

    @property
    def inputs(self) -> Dict[str, Any]:
        return self._inputs

    @property
    def outputs(self) -> NamedTuple:
        return self._outputs

    @property
    def runtime(self) -> float:
        return self._runtime

    @property
    def op_key(self) -> Tuple:
        """Calculates a unique key given the operator hub ID and the operator's
        initialization arguments. Operators with the same hub ID but different
        initialization arguments are considered separate.

        Returns:
            (`tuple`)
                Unique value corresponding to a combination of operator ininitialization
                arguments and its operator ID in the hub.
        """
        args_tup = tuple((key, self.op_args[key])
                         for key in sorted(self.op_args))
        return (self.hub_op_id, ) + args_tup

    def add_ready_handlers(self, handlers):
        """Adds a ready handler to this `Task` object.

        Args:
            handlers: (`typing.Callable`, or `list` of `typing.Callable`)
                Handlers that will be called once the task is ready to be executed.
        """
        if isinstance(handlers, list):
            self._on_ready_handlers += handlers
        else:
            self._on_ready_handlers.append(handlers)

    def add_start_handler(self, handler):
        """Adds a start handler to this `Task` object.

        Args:
            handler: (`typing.Callable`, or `list` of `typing.Callable`)
                Handler that will be called prior to execution.
        """
        if isinstance(handler, list):
            self._on_start_handlers += handler
        else:
            self._on_start_handlers.append(handler)

    def add_finish_handler(self, handler):
        """Adds a finish handler to this `Task` object.

        Args:
            handler: (`typing.Callable`, or `list` of `typing.Callable`)
                Handler that will be called when the `outputs` attribute is set.
        """
        if isinstance(handler, list):
            self._on_finish_handlers += handler
        else:
            self._on_finish_handlers.append(handler)

    def _execute_handlers(self, handlers: List[Callable]):
        """Execute handlers defined in `handlers`. These should be a list of callables
        which take this `Task` object as its input.

        Args:
            handlers: (`list[typing.Callable]`)
                A list of functions to be executed. Should be one of
                `_on_ready_handlers`, `_on_start_handlers`, and `_on_finish_handlers`.
        """
        for handler in handlers:
            handler(self)

    def execute(self, op: Operator):
        """Given a corresponding `Operator` from the `TaskExecutor`, run the task.
        """
        self._execute_handlers(self._on_start_handlers)
        start = timeit.default_timer()

        # Run the operator. The graph provided to the engine should already have done
        # graph validity checks, so further input/output argument checks are
        # unnecessary.
        try:
            self._outputs = op(**self._inputs)
        except Exception as e:  # pylint: disable=broad-except
            logging.error(e)
            self._outputs = None

        self._runtime = timeit.default_timer() - start
        self._execute_handlers(self._on_finish_handlers)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest


from towhee.tests.mock_operators import ADD_OPERATOR_PATH, load_local_operator
from towhee.operator import SharedType


class TestOperator(unittest.TestCase):
    """
    Simple operator test
    """

    def test_func_operator(self):
        add_operator = load_local_operator(
            'add_operator', ADD_OPERATOR_PATH)
        op = add_operator.AddOperator(1)
        self.assertEqual(op(1).sum, 2)
        self.assertEqual(op.shared_type, SharedType.Shareable)

        op2 = add_operator.AddOperator(3)
        self.assertEqual(op2(1).sum, 4)
        self.assertEqual(op2.shared_type, SharedType.Shareable)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from towhee.dag.base_repr import BaseRepr


class OperatorRepr(BaseRepr):
    """This class encapsulates operator representations at compile-time.

    Args:
        name:
            Name of the operator represented by this object.
        df_in:
            Input dataframes(s) to this object.
        df_out:
            This operator's output dataframe.
        iterator:
            This operator's iterator info.
    """
    def __init__(
        self,
        name: str,
        function: str,
        init_args: Dict[str, any],
        inputs: List[Dict[str, any]],
        outputs: List[Dict[str, any]],
        iter_info: Dict[str, any]
    ):
        super().__init__(name)
        self._function = function
        self._inputs = inputs
        self._outputs = outputs
        self._init_args = init_args
        self._iter_info = iter_info

    @property
    def function(self):
        return self._function

    @property
    def inputs(self) -> List:
        """
        Returns:
        [
            {
                'name': `arg name`,
                'df': `dataframe name`,
                'col': `col index`
            },
            ...
        ]
        """
        return self._inputs

    @property
    def outputs(self) -> List:
        """
        Returns:
        [
            {
                "df": `dataframe name`
            },
            ...
        ]
        """
        return self._outputs

    @property
    def init_args(self) -> Dict:
        """
        Returns:
        {
            `arg_name`: `arg value`,
            ...
        }
        """
        return self._init_args

    @property
    def iter_info(self):
        """
        Returns:
        {
            "type": `iter type`
        }
        """
        return self._iter_info

    @staticmethod
    def from_dict(info: Dict) -> 'OperatorRepr':
        """
        Args:
            info:
               Operator info
               {
                   'name': 'example_op',
                   'function': 'test_function',
                   'init_args': {
                       'arg1': 1,
                       'arg2': 'test'
                   },
                   'inputs': [
                       {
                           'df': 'input1',
                           'col': 0
                       },
                       {
                           'df': 'input2',
                           'col': 0
                       }
                   ],
                   'outputs': [
                       {
                           'df': 'input1'
                       },
                   ],
                   'iter_info': {
                       'type': 'map'
                   }
               }

            dataframes(`dict`):
                Dict with `DataFrameRepr` objects as values to construct current
                `OperatorRepr`.

        Returns:
            The operators we described in `file_or_src`
        """
        if not BaseRepr.is_valid(info, {'name', 'init_args', 'function', 'inputs', 'outputs', 'iter_info'}):
            raise ValueError('Invalid operator info.')

        return OperatorRepr(info['name'], info['function'], info['init_args'], info['inputs'], info['outputs'], info['iter_info'])



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Scheduler utilities for pytorch optimization."""


import math
from torch.optim.lr_scheduler import LambdaLR
from torch.optim.optimizer import Optimizer


scheduler_list = [
    'constant',
    'constant_with_warmup',
    'linear',
    'cosine',
    'cosine_with_restarts',
    'polynomial'
]


def check_scheduler(scheduler_type: str) -> bool:
    """
    Check if the scheduler type is supported.

    Args:
        scheduler_type:
            the type of the scheduler.
    Return:
        if the scheduler type is supported.
    """
    if scheduler_list.count(scheduler_type) == 0:
        return False
    else:
        return True


def configure_constant_scheduler(optimizer: Optimizer, last_epoch: int = -1):
    """
    Return a scheduler with a constant learning rate, using the learning rate set in optimizer.

    Args:
        optimizer:
            The optimizer for which to schedule the learning rate.
        last_epoch:
            The last epoch when resuming training.

    Return:
        A constant scheduler.
    """
    return LambdaLR(optimizer, lambda _: 1, last_epoch=last_epoch)


def configure_constant_scheduler_with_warmup(optimizer: Optimizer, num_warmup_steps: int, last_epoch: int = -1):
    """
    Return a schedule with a constant learning rate preceded by a warmup period during which the learning rate
    increases linearly between 0 and the initial lr set in the optimizer.

    Args:
        optimizer:
            The optimizer to be scheduled.
        num_warmup_steps:
            Warmup steps.
        last_epoch:
            The last epoch when training is resumed.

    Return:
        A constant scheduler with warmup.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1.0, num_warmup_steps))
        return 1.0

    return LambdaLR(optimizer, lr_lambda, last_epoch=last_epoch)


def configure_linear_scheduler_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    """
    Return a scheduler with a learning rate that decreases linearly from the initial lr set in the optimizer to 0, after
    a warmup period during which it increases linearly from 0 to the initial lr set in the optimizer.

    Args:
        optimizer:
            The optimizer to be scheduled.
        num_warmup_steps:
            Warmup steps.
        num_training_steps:
            Training steps.
        last_epoch:
            The last epoch when training is resumed.

    Return:
        A linear scheduler with warmup.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def configure_cosine_scheduler_with_warmup(
    optimizer: Optimizer, num_warmup_steps: int, num_training_steps: int,num_cycles: float = 0.5, last_epoch: int = -1
):
    """
    Return a scheduler with a learning rate that decreases following the values of the cosine function between the
    initial lr set in the optimizer to 0, after a warmup period during which it increases linearly between 0 and the
    initial lr set in the optimizer.

    Args:
        optimizer:
            The optimizer to be scheduled.
        num_warmup_steps:
            The steps for the warmup phase.
        num_training_steps:
            The number of training steps.
        num_cycles:
            The number of periods in te cosine scheduler.
        last_epoch:
            The last epoch when training is resumed.

    Return:
        A cosine scheduler with warmup.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def configure_cosine_with_hard_restarts_scheduler_with_warmup(
    optimizer: Optimizer, num_warmup_steps: int, num_training_steps: int, num_cycles: int = 1, last_epoch: int = -1
):
    """
    Return a scheduler with a learning rate that decreases following the values of the cosine function between the
    initial lr set in the optimizer to 0, with several hard restarts, after a warmup period during which it increases
    linearly between 0 and the initial lr set in the optimizer.

    Args:
        optimizer:
            The optimizer to be scheduled.
        num_warmup_steps:
            The steps for the warmup phase.
        num_training_steps:
            The number of training steps.
        num_cycles:
            The number of hard restarts to be used.
        last_epoch:
            The index of the last epoch when training is resumed.

    Return:
        A cosine with hard restarts scheduler with warmup.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        if progress >= 1.0:
            return 0.0
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * ((float(num_cycles) * progress) % 1.0))))

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def configure_polynomial_decay_scheduler_with_warmup(
    optimizer, num_warmup_steps, num_training_steps, lr_end=1e-7, power=1.0, last_epoch=-1
):
    """
    Return a scheduler with a learning rate that decreases as a polynomial decay from the initial lr set in the
    optimizer to end lr defined by `lr_end`, after a warmup period during which it increases linearly from 0 to the
    initial lr set in the optimizer.

    Args:
        optimizer:
            The optimizer to be scheduled.
        num_warmup_steps:
            The steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The number of training steps
        lr_end:
            The end LR.
        power:
            Power factor.
        last_epoch:
            The index of the last epoch when training is resumed.

    Return:
        A polynomial decay scheduler with warmup.
    """

    lr_init = optimizer.defaults['lr']
    assert lr_init > lr_end, f'lr_end ({lr_end}) must be be smaller than initial lr ({lr_init})'

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        elif current_step > num_training_steps:
            return lr_end / lr_init  # as LambdaLR multiplies by lr_init
        else:
            lr_range = lr_init - lr_end
            decay_steps = num_training_steps - num_warmup_steps
            pct_remaining = 1 - (current_step - num_warmup_steps) / decay_steps
            decay = lr_range * pct_remaining ** power + lr_end
            return decay / lr_init  # as LambdaLR multiplies by lr_init

    return LambdaLR(optimizer, lr_lambda, last_epoch)


# class AdafactorScheduler(LambdaLR):
#     """
#     Adafactor scheduler.
#
#     It returns ``initial_lr`` during startup and the actual ``lr`` during stepping.
#     """
#
#     def __init__(self, optimizer: Optimizer, initial_lr=0.0):
#         def lr_lambda(_):
#             return initial_lr
#
#         for group in optimizer.param_groups:
#             group['initial_lr'] = initial_lr
#         super().__init__(optimizer, lr_lambda)
#         for group in optimizer.param_groups:
#             del group['initial_lr']
#
#     def get_lr(self):
#         opt = self.optimizer
#         lrs = [
#             opt._get_lr(group, opt.state[group['params'][0]])
#             for group in opt.param_groups
#             if group['params'][0].grad is not None
#         ]
#         if len(lrs) == 0:
#             lrs = self.base_lrs  # if called before stepping
#         return lrs
#
#
# def congigure_adafactor_scheduler(optimizer: Optimizer, initial_lr=0.0):
#     """
#     Return an Adafactor scheduler.
#
#     Args:
#         optimizer:
#             The optimizer to be scheduled.
#         initial_lr:
#             Initial lr.
#
#     Return:
#         An Adafactor scheduler.
#     """
#     return AdafactorScheduler(optimizer, initial_lr)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT_ WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os
from towhee.dag.utils.callstack import Callstack

cur_dir = os.path.abspath(__file__)


def collect_frames(ignore: int = 0):
    stack = Callstack(ignore)
    stack.frames = [f for f in stack.frames if f.filename == cur_dir]
    stack.size = len(stack.frames)
    return stack


def generate_callstack(ignore: int = 0):
    return collect_frames(ignore)


# create a callstack contains 3 frames
# the structure of the callstack (from top to bottom):
# `collect_frames` frame, `generate_callstack` frame, `<module>` frame
s1 = generate_callstack()
# create a callstack contains 2 frames
s2 = generate_callstack(ignore=1)
# create a callstack contains 1 frames
s3 = generate_callstack(ignore=2)
# create a callstack contains 0 frames
s4 = generate_callstack(ignore=3)


class TestCallStack(unittest.TestCase):
    """In this class, we test the functions in Callstack class.
    We do not initialize the Callstack directly. Instead, we create two functions
    `generate_frames` and `collect_frames` to initialize the Callstack as well as
    block the frames produced by the unittest. We do not care about the process of
    tests, we want to focus only on the architectures that we designed, which in this
    case are `test_xxx`, `generate_callstack` and `collect_frames`. In this way, the
    structure and depth of the callstack is stable and predictable.
    For convenience, we collect and returns the frames inside the callsatck in the
    the form of list.
    """
    def test_init(self):
        # the callstacks are instances of Callstack class
        self.assertIsInstance(s1, Callstack)
        self.assertIsInstance(s2, Callstack)
        self.assertIsInstance(s3, Callstack)
        self.assertIsInstance(s4, Callstack)

        # the frames should be stored in a list
        self.assertIsInstance(s1.frames, list)
        self.assertIsInstance(s2.frames, list)
        self.assertIsInstance(s3.frames, list)
        self.assertIsInstance(s4.frames, list)

    def test_num_frames(self):
        # assert the size of the callstacks, namely 3, 2, 1, 0
        self.assertEqual(s1.num_frames(), 3)
        self.assertEqual(s2.num_frames(), 2)
        self.assertEqual(s3.num_frames(), 1)
        self.assertEqual(s4.num_frames(), 0)

    def test_find_func(self):
        # create a callstack contain 3 frames
        s = generate_callstack()

        # the `test_find_func` is the first function to be called in this case
        # so it's in the bottom of the current callstack, the index is 2
        self.assertEqual(s.find_func('test_find_func'), 2)
        # the `generate_callstack` is the secondly called function
        self.assertEqual(s.find_func('generate_callstack'), 1)
        # the `collect_frames` is the last function
        self.assertEqual(s.find_func('collect_frames'), 0)
        # the function `func_that_not_exist` does not exist in current callstack
        # therefore return None
        self.assertIsNone(s.find_func('func_that_not_exist'))

    def test_hash(self):
        # create a callstack contain 3 frames
        s = generate_callstack()

        # `out_1` and `out_2` have the same range of frames and items
        # hence the same hash value
        out_1 = s.hash(items=['code_context'])
        out_2 = s.hash(items=['code_context'])
        self.assertEqual(out_1, out_2)

        # `out_3` and `out_3` have the same range of frames and items
        # hence the same hash value
        out_3 = s.hash(0, 2, ['filename', 'function', 'position'])
        out_4 = s.hash(0, 2, ['filename', 'function', 'position'])
        self.assertEqual(out_3, out_4)

        # though `out_5` and `out_6` should be different, the lengths are the same
        out_5 = s.hash(items=['filename', 'lineno', 'function', 'code_context', 'position', 'lasti'])
        out_6 = s.hash(0, 1, ['filename'])
        self.assertEqual(len(out_5), len(out_6))

        # raise ValueError when given unsupported item
        self.assertRaises(ValueError, s.hash, items=['attribute_that_not_exists'])
        # raise IndexError when give invalid range of frames
        self.assertRaises(IndexError, s.hash, -1, 1, items=['filename'])
        self.assertRaises(IndexError, s.hash, 0, 1000, items=['filename'])
        self.assertRaises(IndexError, s.hash, 10, 1, items=['filename'])


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from towhee.utils import (
    HandlerMixin,
)


class TestHandlerMixin(unittest.TestCase):
    """Basic test case for `HandlerMixin`.
    """

    class CallbackAdd:
        def __init__(self):
            self.y = 0

        def add(self, x):
            self.y += x

    class CallbackSub:
        def __init__(self):
            self.y = 0

        def sub(self, x):
            self.y -= x

    class Foo(HandlerMixin):
        def __init__(self, x):
            self.add_handler_methods('cal')
            self._x = x

        def call_handlers(self):
            self.call_cal_handlers(self._x)

    def test_handler_add_and_call(self):

        foo = self.Foo(5)

        callback_add = self.CallbackAdd()
        callback_sub = self.CallbackSub()

        foo.add_cal_handler(callback_add.add)
        foo.add_cal_handler(callback_sub.sub)

        foo.call_handlers()

        self.assertEqual(callback_add.y, 5)
        self.assertEqual(callback_sub.y, -5)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Dict

from towhee.dag.base_repr import BaseRepr
from towhee.dag.variable_repr import VariableRepr


class DataFrameRepr(BaseRepr):
    """`DataFrameRepr` represents a single dataframe within a graph.

    A single dataframe is composed of multiple individual variables, each of which is
    required in the next operator stage within the graph.

    Args:
        name:
            The representation name.
        src:
            The information of this dataframe.
    """
    def __init__(self, name: str, columns: List[VariableRepr]):
        self._name = name
        self._columns = columns

    @property
    def name(self):
        return self._name

    @property
    def columns(self):
        return self._columns

    @staticmethod
    def from_dict(info: Dict[str, any]) -> 'DataFrameRepr':
        """
        info:
            dataframe info
            {
                'name': 'example_df',
                'columns': [
                    {'vtype': 'int'},
                    {'vtype': 'float'},
                ],
            }
        Returns:
            DataFrameRepr obj
        """
        if not DataFrameRepr.is_valid(info, {'name', 'columns'}):
            raise ValueError('Invalid dataframe info.')
        var_reprs = []
        df_name = info['name']
        for index, col in enumerate(info['columns']):
            var_reprs.append(VariableRepr(col['name'], col['vtype'], f'{df_name}_col_{index}'))
        return DataFrameRepr(df_name, var_reprs)



# Copyright 2021 Facebook and Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
import math
from torch.optim import Optimizer


class Adafactor(Optimizer):
    """
    AdaFactor pytorch implementation as introduced in `Adafactor: Adaptive Learning Rates with Sublinear Memory Cost`
    https://arxiv.org/abs/1804.04235.
    Arguments:
        params (Iterable[nn.parameter.Parameter]):
            Iterable of parameters to optimize or dictionaries defining parameter groups.
        lr (float, optional):
            The external learning rate.
        eps (Tuple[float, float], optional):
            Regularization constants for square gradient and parameter scale respectively.
        clip_threshold (float, optional):
            Threshold of root mean square of final gradient update.
        decay_rate (float, optional):
            Coefficient used to compute running averages of square.
        beta1 (float, optional):
            Coefficient used for computing running averages of gradient.
        weight_decay (float, optional):
            Weight decay (L2 penalty).
        scale_parameter (bool, optional):
            If True, learning rate is scaled by root mean square.
        relative_step (bool, optional):
            If True, time-dependent learning rate is computed instead of external learning rate.
        warmup_init (bool, optional):
            Time-dependent learning rate computation depends on whether warm-up initialization is being used.
    """

    def __init__(
        self,
        params,
        lr=None,
        eps=(1e-30, 1e-3),
        clip_threshold=1.0,
        decay_rate=-0.8,
        beta1=None,
        weight_decay=0.0,
        scale_parameter=True,
        relative_step=True,
        warmup_init=False,
    ):
        if lr is not None and relative_step:
            raise ValueError("Cannot combine manual `lr` and `relative_step=True` options")
        if warmup_init and not relative_step:
            raise ValueError("`warmup_init=True` requires `relative_step=True`")

        defaults = dict(
            lr=lr,
            eps=eps,
            clip_threshold=clip_threshold,
            decay_rate=decay_rate,
            beta1=beta1,
            weight_decay=weight_decay,
            scale_parameter=scale_parameter,
            relative_step=relative_step,
            warmup_init=warmup_init,
        )
        super().__init__(params, defaults)

    @staticmethod
    def _get_lr(param_group, param_state):
        rel_step_sz = param_group["lr"]
        if param_group["relative_step"]:
            min_step = 1e-6 * param_state["step"] if param_group["warmup_init"] else 1e-2
            rel_step_sz = min(min_step, 1.0 / math.sqrt(param_state["step"]))
        param_scale = 1.0
        if param_group["scale_parameter"]:
            param_scale = max(param_group["eps"][1], param_state["RMS"])
        return param_scale * rel_step_sz

    @staticmethod
    def _get_options(param_group, param_shape):
        factored = len(param_shape) >= 2
        use_first_moment = param_group["beta1"] is not None
        return factored, use_first_moment

    @staticmethod
    def _rms(tensor):
        return tensor.norm(2) / (tensor.numel() ** 0.5)

    @staticmethod
    def _approx_sq_grad(exp_avg_sq_row, exp_avg_sq_col):
        r_factor = (exp_avg_sq_row / exp_avg_sq_row.mean(dim=-1, keepdim=True)).rsqrt_()
        c_factor = exp_avg_sq_col.rsqrt()
        return torch.mm(r_factor.unsqueeze(-1), c_factor.unsqueeze(0))

    def step(self, closure=None):
        """
        Performs a single optimization step
        Arguments:
            closure(callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.dtype in {torch.float16, torch.bfloat16}:
                    grad = grad.float()
                if grad.is_sparse:
                    raise RuntimeError("Adafactor does not support sparse gradients.")

                state = self.state[p]
                grad_shape = grad.shape

                factored, use_first_moment = self._get_options(group, grad_shape)
                # State Initialization
                if len(state) == 0:
                    state["step"] = 0

                    if use_first_moment:
                        # Exponential moving average of gradient values
                        state["exp_avg"] = torch.zeros_like(grad)
                    if factored:
                        state["exp_avg_sq_row"] = torch.zeros(grad_shape[:-1]).to(grad)
                        state["exp_avg_sq_col"] = torch.zeros(grad_shape[:-2] + grad_shape[-1:]).to(grad)
                    else:
                        state["exp_avg_sq"] = torch.zeros_like(grad)

                    state["RMS"] = 0
                else:
                    if use_first_moment:
                        state["exp_avg"] = state["exp_avg"].to(grad)
                    if factored:
                        state["exp_avg_sq_row"] = state["exp_avg_sq_row"].to(grad)
                        state["exp_avg_sq_col"] = state["exp_avg_sq_col"].to(grad)
                    else:
                        state["exp_avg_sq"] = state["exp_avg_sq"].to(grad)

                p_data_fp32 = p.data
                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p_data_fp32 = p_data_fp32.float()

                state["step"] += 1
                state["RMS"] = self._rms(p_data_fp32)
                lr = self._get_lr(group, state)

                beta2t = 1.0 - math.pow(state["step"], group["decay_rate"])
                update = (grad ** 2) + group["eps"][0]
                if factored:
                    exp_avg_sq_row = state["exp_avg_sq_row"]
                    exp_avg_sq_col = state["exp_avg_sq_col"]

                    exp_avg_sq_row.mul_(beta2t).add_(update.mean(dim=-1), alpha=(1.0 - beta2t))
                    exp_avg_sq_col.mul_(beta2t).add_(update.mean(dim=-2), alpha=(1.0 - beta2t))

                    # Approximation of exponential moving average of square of gradient
                    update = self._approx_sq_grad(exp_avg_sq_row, exp_avg_sq_col)
                    update.mul_(grad)
                else:
                    exp_avg_sq = state["exp_avg_sq"]

                    exp_avg_sq.mul_(beta2t).add_(update, alpha=(1.0 - beta2t))
                    update = exp_avg_sq.rsqrt().mul_(grad)

                update.div_((self._rms(update) / group["clip_threshold"]).clamp_(min=1.0))
                update.mul_(lr)

                if use_first_moment:
                    exp_avg = state["exp_avg"]
                    exp_avg.mul_(group["beta1"]).add_(update, alpha=(1 - group["beta1"]))
                    update = exp_avg

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(p_data_fp32, alpha=(-group["weight_decay"] * lr))

                p_data_fp32.add_(-update)

                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p.data.copy_(p_data_fp32)

        return loss



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import tempfile
import unittest
import torch
from torch import nn

from towhee.trainer.optimization.adafactor import Adafactor
from towhee.trainer.optimization.adamw import AdamW
from towhee.trainer.scheduler import (
    configure_constant_scheduler,
    configure_constant_scheduler_with_warmup,
    configure_linear_scheduler_with_warmup,
    configure_cosine_scheduler_with_warmup,
    configure_cosine_with_hard_restarts_scheduler_with_warmup,
    configure_polynomial_decay_scheduler_with_warmup
)


def unwrap_scheduler(scheduler, num_steps=10):
    lr_sch = []
    for _ in range(num_steps):
        lr_sch.append(scheduler.get_lr()[0])
        scheduler.step()
    return lr_sch


def unwrap_and_save_reload_scheduler(scheduler, num_steps=10):
    lr_sch = []
    for step in range(num_steps):
        lr_sch.append(scheduler.get_lr()[0])
        scheduler.step()
        if step == num_steps // 2:
            with tempfile.TemporaryDirectory() as tmpdirname:
                file_name = os.path.join(tmpdirname, 'schedule.bin')
                torch.save(scheduler.state_dict(), file_name)

                state_dict = torch.load(file_name)
                scheduler.load_state_dict(state_dict)
    return lr_sch


class ScheduleInitTest(unittest.TestCase):
    mdl = nn.Linear(50, 50)
    optimizer = AdamW(mdl.parameters(), lr=10.0)
    num_steps = 10

    def assertListAlmostEqual(self, list1, list2, tol, msg=None):
        self.assertEqual(len(list1), len(list2))
        for a, b in zip(list1, list2):
            self.assertAlmostEqual(a, b, delta=tol, msg=msg)

    def test_schedulers(self):
        common_kwargs = {'num_warmup_steps': 2, 'num_training_steps': 10}
        # schedulers doct format
        # function: (sched_args_dict, expected_learning_rates)
        scheds = {
            configure_constant_scheduler: ({}, [10.0] * self.num_steps),
            configure_constant_scheduler_with_warmup: (
                {'num_warmup_steps': 4},
                [0.0, 2.5, 5.0, 7.5, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0],
            ),
            configure_linear_scheduler_with_warmup: (
                {**common_kwargs},
                [0.0, 5.0, 10.0, 8.75, 7.5, 6.25, 5.0, 3.75, 2.5, 1.25],
            ),
            configure_cosine_scheduler_with_warmup: (
                {**common_kwargs},
                [0.0, 5.0, 10.0, 9.61, 8.53, 6.91, 5.0, 3.08, 1.46, 0.38],
            ),
            configure_cosine_with_hard_restarts_scheduler_with_warmup: (
                {**common_kwargs, 'num_cycles': 2},
                [0.0, 5.0, 10.0, 8.53, 5.0, 1.46, 10.0, 8.53, 5.0, 1.46],
            ),
            configure_polynomial_decay_scheduler_with_warmup: (
                {**common_kwargs, 'power': 2.0, 'lr_end': 1e-7},
                [0.0, 5.0, 10.0, 7.656, 5.625, 3.906, 2.5, 1.406, 0.625, 0.156],
            ),
        }

        for scheduler_func, data in scheds.items():
            kwargs, expected_learning_rates = data

            scheduler = scheduler_func(self.optimizer, **kwargs)
            self.assertEqual(len([scheduler.get_lr()[0]]), 1)
            lr_sch_1 = unwrap_scheduler(scheduler, self.num_steps)
            self.assertListAlmostEqual(
                lr_sch_1,
                expected_learning_rates,
                tol=1e-2,
                msg=f'failed for {scheduler_func} in normal scheduler',
            )

            scheduler = scheduler_func(self.optimizer, **kwargs)
            lr_sch_2 = unwrap_and_save_reload_scheduler(scheduler, self.num_steps)
            self.assertListEqual(lr_sch_1, lr_sch_2, msg=f'failed for {scheduler_func} in save and reload')


class OptimizationTest(unittest.TestCase):
    def assertListAlmostEqual(self, list1, list2, tol):
        self.assertEqual(len(list1), len(list2))
        for a, b in zip(list1, list2):
            self.assertAlmostEqual(a, b, delta=tol)

    def test_adam_w(self):
        w = torch.tensor([0.1, -0.2, -0.1], requires_grad=True)
        target = torch.tensor([0.4, 0.2, -0.5])
        criterion = nn.MSELoss()
        # No warmup, constant schedule, no gradient clipping
        optimizer = AdamW(params=[w], lr=2e-1, weight_decay=0.0)
        for _ in range(100):
            loss = criterion(w, target)
            loss.backward()
            optimizer.step()
            w.grad.detach_()  # No zero_grad() function on simple tensors. we do it ourselves.
            w.grad.zero_()
        self.assertListAlmostEqual(w.tolist(), [0.4, 0.2, -0.5], tol=1e-2)

    def test_adafactor(self):
        w = torch.tensor([0.1, -0.2, -0.1], requires_grad=True)
        target = torch.tensor([0.4, 0.2, -0.5])
        criterion = nn.MSELoss()
        # No warmup, constant schedule, no gradient clipping
        optimizer = Adafactor(
            params=[w],
            lr=1e-2,
            eps=(1e-30, 1e-3),
            clip_threshold=1.0,
            decay_rate=-0.8,
            beta1=None,
            weight_decay=0.0,
            relative_step=False,
            scale_parameter=False,
            warmup_init=False,
        )
        for _ in range(1000):
            loss = criterion(w, target)
            loss.backward()
            optimizer.step()
            w.grad.detach_()  # No zero_grad() function on simple tensors. we do it ourselves.
            w.grad.zero_()
        self.assertListAlmostEqual(w.tolist(), [0.4, 0.2, -0.5], tol=1e-2)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from collections import defaultdict
import random
import time
from typing import List
import weakref
import threading

from towhee.engine.task import Task
from towhee.engine.task_executor import TaskExecutor


MAX_EXECUTOR_TASKS = 10


class _TaskScheduler(ABC):
    """Task scheduler abstract interface.

    Args:
        task_execs: (`List[towhee.TaskExecutor]`)
            A list of task executors that the `Engine` manages. Also should be
            continuously changing as new executors are acquired.
        max_tasks: (`int`)
            The maximum number of tasks per executor - if the scheduler detects that an
            executor has greater than or equal to than `max_tasks` in queue, it will not
            assign any further tasks to it.
    """

    def __init__(self, task_execs: List[TaskExecutor], max_tasks: int = 10):
        self._task_execs = task_execs
        self._max_tasks = max_tasks
        # self._pipelines = pipelines
        self._graph_ctxs = []
        self._lock = threading.Lock()
        self._need_stop = False

    def stop(self) -> None:
        self._need_stop = True

    def register(self, graph_ctx):
        with self._lock:
            self._graph_ctxs.append(weakref.ref(graph_ctx))

    # def add_pipeline(self, pipeline: Pipeline):
    #     """Add a single pipeline for this scheduler to manage tasks for.

    #     Args:
    #         pipeline: `towhee.Pipeline`
    #             A single pipeline to schedule tasks for.
    #     """
    #     pipeline.add_task_ready_handler(self.on_task_ready)
    #     pipeline.add_task_start_handler(self.on_task_start)
    #     pipeline.add_task_finish_handler(self.on_task_finish)
    #     self._pipelines.append(pipeline)

    def schedule_forever(self, sleep_ms: int = 1000):
        """Runs the a single schedule step in a loop.

        sleep_ms: (`int`)
            Milliseconds to sleep after completing a single scheduling step.
        """
        while not self._need_stop:
            self.schedule_step()
            # TODO(fzliu): compute runtime for bottleneck operator
            time.sleep(sleep_ms / 1000)

    @abstractmethod
    def schedule_step(self):
        raise NotImplementedError

    @abstractmethod
    def on_task_ready(self, task: Task):
        raise NotImplementedError

    @abstractmethod
    def on_task_start(self, task: Task):
        raise NotImplementedError

    @abstractmethod
    def on_task_finish(self, task: Task):
        raise NotImplementedError


class FIFOTaskScheduler(_TaskScheduler):
    """Very basic scheduler that sends tasks to executors in a first-in, first-out
    manner. The scheduler loops through all `OperatorContext` instances within the
    engine, acquiring tasks one at a time if available.

    Args:
        task_execs: (`List[towhee.TaskExecutor]`)
            See `_TaskScheduler` docstring.
    """

    def __init__(self, task_execs: List[TaskExecutor], max_tasks: int = 10):
        super().__init__(task_execs, max_tasks)

        # `FIFOTaskScheduler` maintains a dictionary of `hub_op_id` to
        # `List[TaskExecutor]` mappings, tracking which operator IDs are loaded into
        # which task executors.
        self._op_id_exec_map = defaultdict(list)

    def schedule_step(self):
        """This function loops once through all operator contexts in all graphs, adding
        a single task from individual operator contexts as appropriate. For maximum
        efficiency, we prioritize operators towards the end of the graph first.
        """

        # for pipeline in self._pipelines:
        with self._lock:
            # for graph_ctx in pipeline.graph_contexts:
            self._graph_ctxs = [
                graph_ctx_ref for graph_ctx_ref in self._graph_ctxs if graph_ctx_ref() is not None
            ]
            for graph_context_ref in self._graph_ctxs:
                g = graph_context_ref()
                if g is None:
                    continue

                for _, op_ctx in g.operator_contexts.items():

                    if not op_ctx.is_schedulable or op_ctx.is_finished:
                        continue

                    tasks = op_ctx.pop_ready_tasks(n_tasks=1)
                    if not tasks:
                        continue
                    task = tasks[0]

                    # If `push_task` returns `False`, then the optimal executor is already
                    # full, wait a while and try again.
                    task_exec = self._find_optimal_exec(task)
                    while not task_exec.push_task(task):
                        time.sleep(self._sleep_ms / 1000)
                        task_exec = self._find_optimal_exec(task)

    def _find_optimal_exec(self, task: Task):
        """Acquires the least busy instance of `TaskExecutor` that can still execute
        the operator.

        Args:
            task: (`towhee.Task`)
                The `Task` instance that the executor will need to complete.
        """

        # Attempt to find the least busy executor with the model already loaded. We do
        # not consider executors that have more than `MAX_EXECUTOR_TASKS` queued up.
        min_num_tasks = self._max_tasks
        optimal_exec = None
        for task_exec in self._op_id_exec_map[task.op_key]:
            if task_exec.num_tasks < min_num_tasks:
                optimal_exec = task_exec
                min_num_tasks = task_exec.num_tasks

        # TODO(fzliu): If no task executor with the specified object was found, assign
        # the least busy executor that has the resource (i.e. GPU, TPU, FPGA, etc...)
        # required by the task.
        if not optimal_exec:
            optimal_exec = random.choice(self._task_execs)
            self._op_id_exec_map[task.op_key].append(optimal_exec)

        return optimal_exec

    def on_task_ready(self, task: Task):
        pass

    def on_task_start(self, task: Task):
        pass

    def on_task_finish(self, task: Task):
        pass



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import torch

from towhee.trainer.models.vit.transformer import merge_last


class TestTransformer(unittest.TestCase):
    """
    Test transformer.
    """
    x = torch.tensor([[[1., -1.], [1., -1.]], [[1., -1.], [1., -1.]]])

    def test_transformer(self):
        y = merge_last(self.x, 2)
        z = torch.tensor([[1., -1., 1., -1.], [1., -1., 1., -1.]])
        c = (y.numpy() == z.numpy())
        self.assertEqual(c.all(), True)


if __name__ == '__main__':
    unittest.main()



# Copyright 2020 The HuggingFace Team and 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import socket
from dataclasses import asdict, dataclass, field
from datetime import datetime
from enum import Enum
from typing import List, Optional

import torch

from towhee.cnn_trainer.utils import logging

logger = logging.get_logger(__name__)
log_levels = logging.get_log_levels_dict().copy()
trainer_log_levels = dict(**log_levels, passive=-1)


def default_logdir() -> str:
    """
    Same default as PyTorch
    """

    current_time = datetime.now().strftime("%b%d_%H-%M-%S")
    return os.path.join("runs", current_time + "_" + socket.gethostname())


@dataclass
class TrainingArguments:
    """
        Training arguments of CNN trainer.

        Parameters:
            output_dir (:obj:`str`):
                The output directory where the model predictions and checkpoints will be written.
            overwrite_output_dir (:obj:`bool`, `optional`, defaults to :obj:`False`):
                If :obj:`True`, overwrite the content of the output directory. Use this to continue training if
                :obj:`output_dir` points to a checkpoint directory.
            prediction_loss_only (:obj:`bool`, `optional`, defaults to `False`):
                When performing evaluation and generating predictions, only returns the loss.
            per_device_train_batch_size (:obj:`int`, `optional`, defaults to 8):
                The batch size per GPU core/CPU for training.
            per_device_eval_batch_size (:obj:`int`, `optional`, defaults to 8):
                The batch size per GPU core/CPU for evaluation.
            eval_accumulation_steps (:obj:`int`, `optional`):
                Number of predictions steps to accumulate the output tensors for, before moving the results to the CPU. If
                left unset, the whole predictions are accumulated on GPU before being moved to the CPU (faster but
                requires more memory).
            num_train_epochs(:obj:`float`, `optional`, defaults to 3.0):
                Total number of training epochs to perform (if not an integer, will perform the decimal part percents of
                the last epoch before stopping training).
            max_steps (:obj:`int`, `optional`, defaults to -1):
                If set to a positive number, the total number of training steps to perform. Overrides
                :obj:`num_train_epochs`.
            no_cuda (:obj:`bool`, `optional`, defaults to :obj:`False`):
                Whether to not use CUDA even when it is available or not.
            eval_steps (:obj:`int`, `optional`):
                Number of update steps between two evaluations if :obj:`evaluation_strategy="steps"`.
            disable_tqdm (:obj:`bool`, `optional`):
                Whether or not to disable the tqdm progress bars.
            label_names (:obj:`List[str]`, `optional`):
                The list of keys in your dictionary of inputs that correspond to the labels.
        """

    output_dir: str = field(
        metadata={"help": "The output directory where the model predictions and checkpoints will be written."},
    )
    overwrite_output_dir: bool = field(
        default=False,
        metadata={
            "help": (
                "Overwrite the content of the output directory."
                "Use this to continue training if output_dir points to a checkpoint directory."
            )
        },
    )

    prediction_loss_only: bool = field(
        default=False,
        metadata={"help": "When performing evaluation and predictions, only returns the loss."},
    )

    per_gpu_train_batch_size: Optional[int] = field(
        default=None,
        metadata={
            "help": "Deprecated, the use of `--per_device_train_batch_size` is preferred. "
                    "Batch size per GPU/TPU core/CPU for training."
        },
    )
    per_gpu_eval_batch_size: Optional[int] = field(
        default=None,
        metadata={
            "help": "Deprecated, the use of `--per_device_eval_batch_size` is preferred."
                    "Batch size per GPU/TPU core/CPU for evaluation."
        },
    )

    eval_accumulation_steps: Optional[int] = field(
        default=None,
        metadata={"help": "Number of predictions steps to accumulate before moving the tensors to the CPU."},
    )

    num_train_epochs: float = field(default=3.0, metadata={"help": "Total number of training epochs to perform."})
    max_steps: int = field(
        default=-1,
        metadata={"help": "If > 0: set total number of training steps to perform. Override num_train_epochs."},
    )

    eval_steps: int = field(default=None, metadata={"help": "Run an evaluation every X steps."})

    disable_tqdm: Optional[bool] = field(
        default=None, metadata={"help": "Whether or not to disable the tqdm progress bars."}
    )

    label_names: Optional[List[str]] = field(
        default=None, metadata={"help": "The list of keys in your dictionary of inputs that correspond to the labels."}
    )

    _n_gpu: int = field(init=False, repr=False, default=-1)
    no_cuda: bool = field(default=False, metadata={"help": "Do not use CUDA even when it is available"})

    def __post_init__(self):
        if self.output_dir is not None:
            self.output_dir = os.path.expanduser(self.output_dir)

        # if self.disable_tqdm is None:
        #     self.disable_tqdm = logger.getEffectiveLevel() > logging.WARN

        self.should_save = True

    @property
    def train_batch_size(self) -> int:
        """
        The actual batch size for training.
        """
        if self.per_gpu_train_batch_size:
            logger.warning(
                "Using deprecated `--per_gpu_train_batch_size` argument which will be removed in a future "
                "version."
            )
        per_device_batch_size = self.per_gpu_train_batch_size
        train_batch_size = per_device_batch_size * max(1, self.n_gpu)
        return train_batch_size

    def _setup_devices(self) -> "torch.device":
        logger.info("PyTorch: setting up devices")
        if self.no_cuda:
            device = torch.device("cpu")
            self._n_gpu = 0
        else:
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            self._n_gpu = torch.cuda.device_count()

        return device

    def to_dict(self):
        """
        Serializes this instance while replace `Enum` by their values (for JSON serialization support).
        """
        d = asdict(self)
        for k, v in d.items():
            if isinstance(v, Enum):
                d[k] = v.value
            if isinstance(v, list) and len(v) > 0 and isinstance(v[0], Enum):
                d[k] = [x.value for x in v]
        return d

    def to_json_string(self):
        """
        Serializes this instance to a JSON string.
        """
        return json.dumps(self.to_dict(), indent=2)

    @property
    def n_gpu(self):
        """
        The number of GPUs used by this process.

        Note:
            This will only be greater than one when you have multiple GPUs available but are not using distributed
            training. For distributed training, it will always be 1.
        """
        # Make sure `self._n_gpu` is properly setup.
        _ = self._setup_devices
        return self._n_gpu

    @property
    def device(self) -> "torch.device":
        """
        The device used by this process.
        """
        return self._setup_devices



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# from towhee.engine.pipeline import Pipeline

from typing import Tuple
import threading

from towhee.dataframe import DataFrame
from towhee.dag.graph_repr import GraphRepr


from towhee.engine.operator_context import OperatorContext


class GraphContext:
    """
    `GraphContext` is a data processing network with one or multiple `Operator`.
    Each row of a `Pipeline`'s inputs will be processed individually by a
    `GraphContext`.

    Args:
        ctx_idx: (int)
            The index of this `GraphContext`
        graph_repr: (`towhee.dag.GraphRepr`)
            The DAG representation
    """

    def __init__(self, ctx_idx: int, graph_repr: GraphRepr):
        self._idx = ctx_idx
        self._repr = graph_repr
        self._mutex = threading.Lock()
        self._finished = threading.Condition(self._mutex)

        self.on_start_handlers = []
        self.on_finish_handlers = []

        self.on_task_ready_handlers = []
        self.on_task_start_handlers = []
        # self.on_task_finish_handlers = [self._on_task_finish]
        self.on_task_finish_handlers = []

        self._build()
        self._cv = threading.Condition()

    def __call__(self, inputs: Tuple):
        graph_input = self.operator_contexts['_start_op'].outputs[0]
        graph_input.put(inputs)
        graph_input.seal()

    def result(self):
        output_df = self.operator_contexts['_end_op'].inputs[0]
        output_df.wait_sealed()
        return output_df

    # def __call__(self, inputs: Tuple):
    #     # todo: GuoRentong, issue #114
    #     graph_input = self.operator_contexts['_start_op'].inputs[0]
    #     graph_input.clear()
    #     graph_input.put(inputs)
    #     graph_input.seal()
    #     self._is_busy = True

    @property
    def operator_contexts(self):
        return self._op_contexts

    @property
    def dataframes(self):
        return self._dataframes

    # @property
    # def outputs(self) -> DataFrame:
    #     return self.operator_contexts['_end_op'].outputs[0]

    # @property
    # def is_busy(self):
    #     return self._is_busy

    # def _on_task_finish(self, task: Task):
    #     """
    #     Callback after the execution of a `Task`.
    #     """
    #     # todo: GuoRentong, currently, each `GraphContext` can only process one row
    #     # the following finish condition is ugly :(
    #     if self.operator_contexts[task.op_name].outputs[0] is self.outputs:
    #         self._is_busy = False

    #     if not self._is_busy:
    #         # call on_finish_handlers
    #         for handler in self.on_finish_handlers:
    #             handler(self)

    def _build(self):
        # build dataframes
        dfs = {}
        for df_name, df_repr in self._repr.dataframes.items():
            cols = {}
            for i in range(len(df_repr.columns)):
                col = df_repr.columns[i]
                cols[col.name] = {}
                cols[col.name]['index'] = i
                cols[col.name]['type'] = col.vtype
            dfs[df_name] = DataFrame(df_name, cols)
        self._dataframes = dfs

        # build operator contexts
        self._op_contexts = {}
        for _, op_repr in self._repr.operators.items():
            is_schedulable = self._is_schedulable_op(op_repr)
            op_ctx = OperatorContext(op_repr, self.dataframes, is_schedulable)

            op_ctx.on_task_start_handlers += self.on_task_start_handlers
            op_ctx.on_task_ready_handlers += self.on_task_ready_handlers
            op_ctx.on_task_finish_handlers += self.on_task_finish_handlers

            self._op_contexts[op_ctx.name] = op_ctx

    @staticmethod
    def _is_schedulable_op(op_repr):
        op_name = op_repr.name
        if op_name in ('_start_op', '_end_op'):
            return False
        else:
            return True



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#from typing import List, Dict, Tuple
#
#
#from towhee.dag.graph_repr import GraphRepr
#from towhee.dag.operator_repr import OperatorRepr
#from towhee.dataframe import DataFrame
#from towhee.engine.graph_context import GraphContext
#from towhee.engine.operator_context import OperatorContext

#
# class _ReprToCtx:
#    """
#    Representation to graph_ctx, op_ctx, dataframes
#    """
#
#    def __init__(self, graph_yaml: str):
#        self._graph_repr = GraphRepr.from_yaml(graph_yaml)
#        self._dataframes = None
#        self._graph_ctx = None
#        self._op_ctxs = None
#
#    @property
#    def dataframes(self) -> Dict[str, DataFrame]:
#        if self._dataframes is None:
#            frames = {}
#            for df_repr in self._graph_repr.dataframes:
#                cols = {}
#                for i in range(len(df_repr.columns)):
#                    col = df_repr.columns[i]
#                    cols[col.name] = {}
#                    cols[col.name]['index'] = i
#                    cols[col.name]['type'] = col.vtype
#                frames[df_repr.name] = DataFrame(df_repr.name, cols)
#            self._dataframes = frames
#        return self._dataframes
#
#    @property
#    def op_ctxs(self) -> List[OperatorContext]:
#        if self._op_ctxs is None:
#            op_ctxs = []
#            for op_repr in self._graph_repr.operators:
#                op_ctxs.append(self._build_op_ctx(op_repr))
#            self._op_ctxs = op_ctxs
#        return self._op_ctxs
#
#    @property
#    def graph_ctx(self) -> GraphContext:
#        if self._graph_ctx is None:
#            self._graph_ctx = GraphContext(self.op_ctxs)
#        return self._graph_ctx
#
#    def _build_op_ctx(self, op_repr: OperatorRepr) -> OperatorContext:
#        df_in = [self.dataframes[df_info['df']] for df_info in op_repr.inputs]
#        df_out = [self.dataframes[df_info['df']]
#                  for df_info in op_repr.outputs]
#
#        return OperatorContext(
#            OpInfo(op_repr.name, op_repr.function, op_repr.init_args,
#                   op_repr.iter_info['type'],
#                   dict((item['name'], item['col']) for item in op_repr.inputs)),
#            df_in,
#            df_out
#        )
#
#
# def create_ctxs(graph_yaml) -> Tuple[GraphContext, List[DataFrame]]:
#    obj = _ReprToCtx(graph_yaml)
#    return obj.graph_ctx, obj.dataframes
#



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading

from towhee.engine.singleton import singleton
from towhee.engine.pipeline import Pipeline
from towhee.engine.task_scheduler import FIFOTaskScheduler
from towhee.engine.task_executor import TaskExecutor


@singleton
class EngineConfig:
    """
    Global engine config
    """

    def __init__(self):
        self._sched_type = 'fifo'
        self._cache_path = None
        self._sched_interval_ms = 20

    @property
    def sched_type(self):
        return self._sched_type

    @sched_type.setter
    def sched_type(self, sched_type: str):
        self._sched_type = sched_type

    @property
    def cache_path(self):
        return self._cache_path

    @cache_path.setter
    def cache_path(self, cache_path: str):
        self._cache_path = cache_path

    @property
    def sched_interval_ms(self):
        return self._sched_interval_ms

    @sched_interval_ms.setter
    def sched_interval_ms(self, sched_interval_ms: int):
        self._sched_interval_ms = sched_interval_ms

    def __str__(self):
        return str(self.__dict__)


@singleton
class Engine(threading.Thread):
    """Engines are the core component responsible for deliving results to the user. A
    single engine may be composed of multiple pipelines.
    """

    def __init__(self):
        super().__init__()

        self._config = EngineConfig()
        self._pipelines = []
        self.setDaemon(True)

        # Setup executors and scheduler.
        self._setup_task_execs()
        self._setup_task_sched()

    def stop(self) -> None:
        self._task_sched.stop()
        for task_exec in self._task_execs:
            task_exec.stop()

    def run(self):
        self._task_sched.schedule_forever(self._config.sched_interval_ms)

    def add_pipeline(self, pipeline: Pipeline):
        """Add a single pipeline to this engine. Pipelines can be added long after an
        engine has been instantiated.

        Args:
            pipeline: `towhee.Pipeline`
                A single pipeline with which this engine will be assume execution
                responsibility.
        """
        # pipeline.add_task_ready_handler(self._on_task_ready)
        # pipeline.add_task_start_handler(self._on_task_start)
        # pipeline.add_task_finish_handler(self._on_task_finish)

        pipeline.register(self._task_sched)
        self._pipelines.append(pipeline)

    def _setup_task_execs(self):
        """(Initialization function) Scan for devices and create TaskExecutors to
        manage task execution on CPU, GPU, and other devices.
        """
        self._task_execs = []

        # TODO(fzliu): Perform device scan.
        dev_names = ['cpu:0']

        # Create executor threads and begin running.
        for name in dev_names:
            executor = TaskExecutor(
                name=name, cache_path=self._config.cache_path)
            self._task_execs.append(executor)
            executor.start()

    def _setup_task_sched(self):
        """(Initialization function) Create a `TaskScheduler` instance.
        """
        self._task_sched = None

        # Parse scheduler type from configuration.
        sched_type = self._config.sched_type
        if sched_type == 'fifo':
            # self._task_sched = FIFOTaskScheduler(
            #     self._pipelines, self._task_execs)
            self._task_sched = FIFOTaskScheduler(self._task_execs)
        else:
            raise ValueError(f'Invalid scheduler type - {sched_type}')

    # def _on_task_ready(self):
    #     """Contains `Engine`-specific code blocks that need to be executed when a task
    #     is marked as ready.
    #     """
    #     pass

    # def _on_task_start(self):
    #     """Contains `Engine`-specific code blocks that need to be executed when a task
    #     begins running.
    #     """
    #     pass

    # def _on_task_finish(self):
    #     """Contains `Engine`-specific code blocks that need to be executed when a task
    #     is finished executing.
    #     """
    #     pass


_engine_lock = threading.Lock()


def start_engine():
    engine = Engine()
    if engine.is_alive():
        return

    with _engine_lock:
        if engine.is_alive():
            return
        engine.start()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import queue

from towhee.engine.task import Task


class TaskQueue:
    """
    A queue for `Task` objects exposed to the `TaskScheduler` which allows the scheduler
    to assign tasks to individual instances of `TaskExecutor`.

    Args:
        size:
            Maximum number of allowable queue elements.
    """

    def __init__(self, size: int = 0):
        self._queue = queue.Queue(size)

    @property
    def empty(self) -> bool:
        """Indicates whether `TaskQueue` is empty. Returns true if the queue has no
        tasks.
        """
        return self._queue.empty()

    @property
    def full(self) -> bool:
        """Indicates whether or not the `TaskQueue` is at its maximum capacity.
        """
        return self._queue.full()

    @property
    def size(self) -> int:
        """Returns the number of tasks in the TaskQueue.
        """
        return self._queue.qsize()

    def push(self, task: Task) -> bool:
        """Pushes a `Task` object to the end of the queue. Returns `True` if the
        operation was successful and `False` otherwise. A return value of `False` most
        likely indicates that the queue has reached its maximum capacity.

        Args:
            task: (`towhee.engine.Task`)
                `Task` object to add to the end of the queue.
        """
        try:
            self._queue.put_nowait(task)
        except queue.Full:
            return False
        return True

    def pop(self) -> Task:
        """Attempts to acquire the first item off of the queue.

        Returns:
            (`towhee.engine.Task`)
                First `Task` object available on the queue, or None if the queue is
                empty
        """
        try:
            return self._queue.get_nowait()
        except queue.Empty:
            return None



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import abstractmethod
from abc import ABC
from enum import Enum


# NotShareable:
#    Stateful operator

# Shareable:
#    Stateless operator
SharedType = Enum('SharedType', ('NotShareable', 'Shareable'))


class Operator(ABC):
    """
    Operator base class, implements __init__ and __call__,

    Examples:
        class AddOperator(Operator):
            def __init__(self, factor: int):
                self._factor = factor

            def __call__(self, num) -> NamedTuple("Outputs", [("sum", int)]):
                Outputs = NamedTuple("Outputs", [("sum", int)])
                return Outputs(self._factor + num)
    """

    @abstractmethod
    def __init__(self):
        """
        Init operator, before a graph starts, the framework will call Operator __init__ function.

        Args:

        Returns:
            None

        Raises:
            An exception during __init__ can terminate the graph run.

        """
        self._key = ''

    @abstractmethod
    def __call__(self):
        """
        The framework calls __call__ function repeatedly for every input data.

        Args:


        Return:
            NamedTuple

        Raises:
            An exception during __init__ can terminate the graph run.
        """
        raise NotImplementedError

    @property
    def key(self):
        return self._key

    @property
    def shared_type(self):
        return SharedType.NotShared

    @key.setter
    def key(self, value):
        self._key = value


class PyTorchNNOperator(Operator):
    """
    PyTorch model operator base
    """

    def train(self):
        """
        For training model
        """
        raise NotImplementedError



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import List, Optional
import hashlib


class Callstack:
    """A 'Callstack' object contains the working frames at the moment.

        Args:
            `ignore (int)`: The number of frames to ignore on the top of the callstack.
    """
    def __init__(self, ignore: int = 0):

        self.frames = inspect.stack()
        # ignore the frame of Callstack.__init__
        ignore += 1
        if ignore > len(self.frames):
            raise ValueError(f"ignore = {ignore-1} is out of frame range")
        del self.frames[0:ignore]
        self.size = len(self.frames)

    def num_frames(self) -> int:
        """Get the number of frames.

        Returns:
            The size of current stack.
        """
        return self.size

    def find_func(self, func_name: str) -> Optional[int]:
        """Given a function name, find the first-matched and outermost frame from current stack.

        Args:
            `func_name (str)`: The function name to find.

        Returns:
            If at least one matching frame exits, return the first-matched frame index.
            Else, return None.
        """
        for i in range(self.size - 1, -1, -1):
            if self.frames[i].function == func_name:
                return i
        return None

    def hash(self, start: int = None, end: int = None, items: List[str] = None) -> str:
        """Get the hash value of the attributes contained in `items` between index `start`
        and `end` (includes `start`, excludes `end`).

        Args:
            `start (int)`: The index of the start frame.

            `end (int)`: The index of the end frame.

            `items (list[str])`: The items to be hashed. Supported items are
            {filename, lineno, function, code_context, position, lasti}, where
            code_context denotes the current line of code of the context, position
            denotes the frame's index of the callstack, lasti denotes the index of last
            attempted instruction in bytecode.

        Returns:
            The hash value.

        Raises:
            `IndexError`: If the args [`start`, `end`) is out of the frame range or
            `end` less than `start`.

            `ValueError`: If an item in `items` is not supported, i.e. not one of
            {filename, lineno, function, code_context, position, lasti}.
        """
        start = start or 0
        end = end or self.size

        if end > self.size or end <= 0 or start >= self.size or start < 0:
            raise IndexError(f"index range [{start}, {end}) out of frame range" f"[0, {self.size})")
        if start >= end:
            raise IndexError(f"end = {end} is less than or equal to start = {start}")

        full_item = {"filename", "lineno", "function", "code_context", "position", "lasti"}
        if not set(items).issubset(set(full_item)):
            invalid_item = set(items) - (set(items) & full_item)
            raise ValueError(f"{invalid_item} not supported")

        md5 = hashlib.md5()
        for i, frame in enumerate(self.frames[start:end]):
            frame_dict = frame._asdict()
            frame_dict["position"] = i + start
            frame_dict["lasti"] = frame_dict["frame"].f_lasti
            frame_dict["code_context"] = ("".join(frame_dict["code_context"]) if frame_dict["code_context"] else "")
            for item in items:
                md5.update(str(frame_dict[item]).encode("utf-8"))
        return md5.hexdigest()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from towhee.trainer.models.vit.vit_configs import get_b16_config


class VitConfigsTest(unittest.TestCase):
    config = get_b16_config()

    def test_vit_config(self):
        self.assertEqual(self.config['dim'], 768)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Optional


class Variable:
    """A `Variable` is a blob of data which can be one of many data types such as:
    `bool` (scalar), `int` (scalar), `float` (scalar), `tuple` (array), `list` (array),
    `np.ndarray` (array), `string` (misc).

    Args:
        name:
            Variable name; should be identical to its representation counterpart.

        vtype:
            type name: int, float, str, bytes ...
        value:
            variable data

    """

    def __init__(self, vtype: str, value: any, name: Optional[str] = None):
        self._vtype = vtype
        self._value = value
        self._name = name

    @property
    def value(self) -> any:
        return self._value

    @property
    def vtype(self) -> str:
        return self._vtype



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable, Dict, NamedTuple

from towhee.engine.task import Task


class EmulatedPipeline:
    """Dummy pipeline with only fields required by the FIFOTaskScheduler filled in.
    """

    def __init__(self):
        GraphContext = NamedTuple('GraphContext', [('operator_contexts', Dict)])
        OperatorContext = NamedTuple('OperatorContext', [('pop_ready_tasks', Callable), ('is_schedulable', Callable)])

        self._on_ready_handlers = []
        self._on_start_handlers = []
        self._on_finish_handlers = []
        self._task_idx = 0

        # Create dummy GraphContext with a single OperatorContext instance.
        self.graph_contexts = [GraphContext({'op': OperatorContext(self._pop_ready_tasks, self._is_schedulable)})]

    def _add_task_handlers(self, task: Task):
        """Helper function which adds handlers to the specified task.
        """
        for handler in self._on_ready_handlers:
            task.add_ready_handler(handler)
        for handler in self._on_start_handlers:
            task.add_start_handler(handler)
        for handler in self._on_finish_handlers:
            task.add_finish_handler(handler)

    def _pop_ready_tasks(self, n_tasks: int = 1):
        """Returns a list of of `add` tasks with continuously increasing `task_idx`.
        """
        hub_op_id = 'mock_operators/add_operator'
        args = {'factor': 0}
        tasks = []
        for n in range(n_tasks):
            num = self._task_idx + n
            task = Task('test', hub_op_id, args, {'num': num}, num)
            self._add_task_handlers(task)
            tasks.append(task)
        self._task_idx += n_tasks
        return tasks

    def _is_schedulable(self):
        return True

    def add_task_ready_handler(self, function: Callable):
        self._on_ready_handlers.append(function)

    def add_task_start_handler(self, function: Callable):
        self._on_start_handlers.append(function)

    def add_task_finish_handler(self, function: Callable):
        self._on_finish_handlers.append(function)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.dag.dataframe_repr import DataFrameRepr
from towhee.dag.graph_repr import GraphRepr
from towhee.dag.operator_repr import OperatorRepr
from towhee.dag.variable_repr import VariableRepr


__all__ = [
    'DataFrameRepr',
    'GraphRepr',
    'OperatorRepr',
    'VariableRepr'
]



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib.util
from pathlib import Path
from typing import Any, Dict

from towhee.operator import Operator
from towhee.engine import LOCAL_OPERATOR_CACHE


class OperatorLoader:
    """Wrapper class used to load operators from either local cache or a remote
    location.

    Args:
        cache_path: (`str`)
            Local cache path to use. If not specified, it will default to
            `$HOME/.towhee/operators`.
    """

    def __init__(self, cache_path: str = None):
        if cache_path is None:
            self._cache_path = LOCAL_OPERATOR_CACHE
        else:
            self._cache_path = Path(cache_path)

    def load_operator(self, function: str, args: Dict[str, Any]) -> Operator:
        """Attempts to load an operator from cache. If it does not exist, looks up the
        operator in a remote location and downloads it to cache instead. By standard
        convention, the operator must be called `Operator` and all associated data must
        be contained within a single directory.

        Args:
            function: (`str`)
                Origin and method/class name of the operator. Used to look up the proper
                operator in cache.
        """

        # Lookup the path for the operator in local cache. Example directory structure:
        #  /home/user/.towhee/operators/organization-name/operator-name
        #   |_ /home/user/.towhee/operators/organization-name/operator-name/operator.py
        #   |_ /home/user/.towhee/operators/organization-name/operator-name/resnet.torch
        #   |_ /home/user/.towhee/operators/organization-name/operator-name/config.json
        fname = function.split('/')[-1]
        path = self._cache_path / function / (fname + '.py')
        # If the following check passes, the desired operator was found locally and can
        # be loaded from cache.
        if path.is_file():

            # Specify the module name and absolute path of the file that needs to be
            # imported.
            modname = 'towhee.operator.' + fname
            spec = importlib.util.spec_from_file_location(
                modname, path.resolve())

            # Create the module and then execute the module in its own namespace.
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Instantiate the operator object and return it to the caller for
            # `load_operator`. By convention, the operator class is simply the CamelCase
            # version of the snake_case operator.
            op_cls = ''.join(x.capitalize() or '_' for x in fname.split('_'))
            return getattr(module, op_cls)(**args)

        else:
            raise FileNotFoundError('Operator definition not found')



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import time
import queue

from towhee.dataframe import DataFrame, DataFrameIterator

from towhee.tests.test_util.dataframe_test_util import DfWriter, MultiThreadRunner


class TestDataframe(unittest.TestCase):
    """
    Test dataframe basic function
    """

    def test_put(self):
        df = DataFrame('test')
        t = DfWriter(df, 10)
        t.start()
        t.join()
        self.assertEqual(df.name, 'test')
        self.assertEqual(df.size, 10)
        end, datas = df.get(0, 4)
        self.assertFalse(end)
        self.assertEqual(len(datas), 4)

        end, datas = df.get(8, 4)
        self.assertFalse(end)
        self.assertEqual(len(datas), 2)

        end, datas = df.get(8, 4, True)
        self.assertFalse(end)
        self.assertEqual(len(datas), 0)

        self.assertFalse(df.is_sealed())
        df.seal()
        self.assertTrue(df.is_sealed())

        end, datas = df.get(8, 4, True)
        self.assertTrue(end)
        self.assertEqual(len(datas), 2)

    def test_multithread(self):
        df = DataFrame('test')
        data_size = 10
        t = DfWriter(df, data_size)
        t.set_sealed_when_stop()
        t.start()
        q = queue.Queue()

        def read(df: DataFrame, q: queue.Queue):
            index = 0
            while True:
                end, items = df.get(index, 2)
                if items:
                    for item in items:
                        q.put(item)
                        index += 1
                if end:
                    break

        runner = MultiThreadRunner(target=read, args=(df, q), thread_num=10)

        runner.start()
        runner.join()

        self.assertEqual(q.qsize(), data_size * 10)


class TestMapIterator(unittest.TestCase):
    """
    map iterator basic test
    """

    def test_map_iterator_multithread(self):
        df = DataFrame('test')
        it = df.map_iter()
        data_size = 100
        t = DfWriter(df, data_size)
        t.start()
        t.set_sealed_when_stop()
        q = queue.Queue()

        def read(it: DataFrameIterator, q: queue.Queue):
            for item in it:
                if item:
                    q.put(item)
                time.sleep(0.01)

        runner = MultiThreadRunner(target=read, args=(it, q), thread_num=5)
        runner.start()
        runner.join()
        self.assertEqual(q.qsize(), 100)

    def test_map_iterator(self):
        df = DataFrame('test')
        it = df.map_iter()
        t = DfWriter(df, 20)
        t.start()
        t.join()

        count = 0
        for item in it:
            if count < 20:
                self.assertEqual(len(item), 1)
                count += 1
                self.assertEqual(it.accessible_size, 20 - count)
            else:
                df.seal()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import importlib
import logging


_MOCK_OPERATOR_DIR = os.path.dirname(__file__)


ADD_OPERATOR_PATH = os.path.join(_MOCK_OPERATOR_DIR, "add_operator")
SUB_OPERATOR_PATH = os.path.join(_MOCK_OPERATOR_DIR, "sub_operator")
PYTORCH_CNN_OPERATOR_PATH = os.path.join(_MOCK_OPERATOR_DIR, "pytorch_cnn_operator")
PYTORCH_TRANSFORMER_OPERATOR_PATH = os.path.join(_MOCK_OPERATOR_DIR, "pytorch_transformer_operator")
PYTORCH_TRANSFORME_OPERATOR_PATH = os.path.join(_MOCK_OPERATOR_DIR, "pytorch_transform_operator")



def load_local_operator(op_name: str, op_path: str):
    """
    A simple operator_loader function for UT
    """
    try:
        sys.path.insert(0, op_path)
        return importlib.import_module(op_name)
    except ModuleNotFoundError as e:
        logging.error(
            "import model: {op_name} from path: {op_path} failed, error: {err}",
            op_name=op_name,
            op_path=op_path,
            err=str(e)
        )
    finally:
        sys.path.pop(0)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np


class Array:
    """One-dimensional array of data

    Args:
        data: (`numpy.ndarray`, `list`, `dict` or scalar)
            Contains data to be stored in `Array`.
        dtype: (`numpy.dtype`, `str`, `PIL.Image` or `towhee.DType`, optional)
            Data type of the `Array`. If not specified, it will be inferred from `data`.
        name: (`str`)
            Name of the `Array`.
        copy: (`bool`, default False)
            Copy the `data` contents to the `Array`.
    """

    def __init__(
        self,
        data=None,
        #dtype=None,
        name: str = None,
        #copy=False
    ):

        self.name = name
        # TODO(GuoRentong): provide towhee.DType abstraction and support
        # Array.dtypes
        self.dtype = None

        # For data is numpy.array
        if isinstance(data, np.ndarray):
            # For `data` is scalar as array
            if data.size == 1 and data.shape == ():
                data = data.reshape((1,))

        # For data is list
        elif isinstance(data, list):
            data = np.array(data)

        # For data is None
        elif data is None:
            # TODO(GuoRentong): remember to handle dtype properly
            pass

        # For data is scalar
        else:
            data = np.array([data])

        self.data = data

    def __iter__(self):
        return self.data.__iter__()

    def __getitem__(self, key: int):
        if isinstance(key, int):
            return self.data[key]
        else:
            raise IndexError("only integers are invalid indices")

    def __repr__(self):
        return self.data.__repr__()

    @property
    def size(self) -> int:
        """ Number of elements in the `Array`.
        """
        return self.data.shape[0]

    def is_empty(self) -> bool:
        """
        Indicator whether Array is empty.
        True if Array has no elements.
        """
        return self.size == 0



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from typing import Callable, List, Tuple, Dict
import weakref

from towhee.dataframe.variable import Variable


class DataFrame:
    """A dataframe is a collection of immutable, potentially heterogeneous blogs of
    data.
    """

    def __init__(self, name: str = None, cols=None, data: List[Tuple[Variable]] = None):
        """DataFrame constructor.

        Args:
            name:
                Name of the dataframe; `DataFrame` names should be the same as its
                representation.
            data:
                A list of data tuples - in all instances, the number of elements per
                tuple should be identical throughout the entire lifetime of the
                `Dataframe`. These tuples can be interpreted as being direct outputs
                into downstream operators.
        """
        self._name = name
        # TODO (junjie.jiangjjj) define col struct
        self._cols = cols
        self._data = data if data else []
        self._registered_iter = []
        self._start_index = 0
        self._total = len(self._data)
        self._sealed = False

        self._lock = threading.Lock()
        self._registered_lock = threading.Lock()
        self._cv = threading.Condition(self._lock)

    @property
    def name(self) -> str:
        return self._name

    @property
    def data(self) -> List:
        return self._data

    def put(self, item: Tuple[Variable]) -> None:
        assert not self._sealed, f'DataFrame {self._name} is already sealed, can not put data'
        assert isinstance(
            item, tuple), 'Dataframe needs tuple, not %s' % (type(item))
        with self._lock:
            self._data.append(item)
            self._total += 1

    def merge(self, df):
        with self._lock:
            self._data += df.data
            self._total += len(df.data)

    def put_dict(self, data: Dict[str, any]):
        datalist = [None] * len(self._cols)
        for k, v in data.items():
            datalist[self._cols[k]['index']] = Variable(
                self._cols[k]['type'], v)
        self.put(tuple(datalist))

    def get(self, start: int, count: int, force_size: bool = False) -> Tuple[bool, List[Tuple[Variable]]]:
        """
        Get [start: start + count) items of dataframe
        `force_size` is true, return data size must equal `count`. If there is
        not enough data, return empty list.
        `force_size` is false, return no more than `count` data

        Returns:
            (is_sealed, data)
        """
        # TODO (junjie.jiang)
        # 1. call gc function
        # 2. put and seal will be called by the same operator
        # 3. seal will always be called after put
        # 4. iterators never ends until they meet df._sealed == True
        with self._lock:
            if start < self._start_index:
                raise IndexError(
                    'Can not read from {start}, dataframe {name} start index is {cur_start}',
                    start=start,
                    name=self.name,
                    cur_start=self._start_index
                )

            if force_size and not self._sealed and self._total - start < count:
                return False, []

            real_count = min(self._total - start, count)
            real_start_index = start - self._start_index
            return self._sealed and real_start_index + real_count >= self._total, \
                self._data[real_start_index: real_start_index + real_count]

    @property
    def size(self):
        return self._total

    def seal(self):
        with self._cv:
            self._sealed = True
            self._cv.notify_all()

    def is_sealed(self) -> bool:
        return self._sealed

    def wait_sealed(self):
        with self._cv:
            if not self._sealed:
                self._cv.wait()

    def clear(self):
        self._data = []
        self._start_index = 0
        self._total = 0
        self._sealed = False

    def _gc(self) -> None:
        # TODO (junjie.jiangjjj)
        """
        Delete the data which all registered iter has read
        """
        return None

    def map_iter(self):
        it = MapIterator(self)
        with self._registered_lock:
            self._registered_iter.append(it)
        return it

    def __str__(self) -> str:
        return 'DataFrame [%s] with [%s] datas, seal: [%s]' % (self.name, self.size, self.is_sealed())


class DataFrameIterator:
    """Base iterator implementation. All iterators should be subclasses of
    `DataFrameIterator`.

    Args:
        df: The dataframe to iterate over.
    """

    def __init__(self, df: DataFrame):
        self._df_ref = weakref.ref(df)
        self._cur_index = 0

    def __iter__(self):
        return self

    def __next__(self):
        # The base iterator is purposely defined to have exatly 0 elements.
        raise StopIteration

    @property
    def accessible_size(self) -> int:
        """
        Return current accessible data size
        """
        return self._df_ref().size - self._cur_index


class MapIterator(DataFrameIterator):
    """Iterator implementation that traverses the dataframe line-by-line.

    Args:
        df: The dataframe to iterate over.
    """

    def __init__(self, df: DataFrame):
        super().__init__(df)
        self._lock = threading.Lock()

    def __next__(self) -> List[Tuple]:
        with self._lock:
            sealed, data = self._df_ref().get(self._cur_index, 1)
            if sealed and len(data) == 0:
                raise StopIteration
            self._cur_index += len(data)
        return data


class BatchIterator:
    """Iterator implementation that traverses the dataframe multiple rows at a time.

    Args:
        df:
            The dataframe to iterate over.
        size:
            batch size. If size is None, then the whole variable set will be batched
            together.
    """

    def __init__(self, df: DataFrame, size: int = None):
        super().__init__(df)
        self._size = size

    def __next__(self):
        raise NotImplementedError


class GroupIterator:
    """Iterator implementation that traverses the dataframe based on a custom group-by
    function.

    Args:
        df:
            The dataframe to iterate over.
        func:
            The function used to return grouped data within the dataframe.
    """

    def __init__(self, df: DataFrame, func: Callable[[], None]):
        super().__init__(df)
        self._func = func

    def __next__(self):
        raise NotImplementedError


class RepeatIterator:
    """Iterator implementation that repeatedly accesses the first line of the dataframe.

        Args:
            df:
                The dataframe to iterate over.
            n:
                the repeat times. If `None`, the iterator will loop continuously.
    """

    def __init__(self, df: DataFrame, n: int = None):
        # self._n = n
        # self._i = 0
        raise NotImplementedError

    def __next__(self):
        if self._i >= self._n:
            raise StopIteration
        return self._data[0]



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple
import torch

from towhee.operator import Operator


class PytorchTransformerOperator(Operator):
    """
    Pytorch transformer operator
        Args:
            model(nn.Module):
                Model object
            args(Optional[Dict]):
                A dict containing all the other arguments, such as labels_map
    """

    def __init__(self, model, args=None) -> None:
        super().__init__()
        self._model = model
        self._args = args

    def __call__(self, img_tensor) -> NamedTuple('Outputs', [('predict', str)]):
        self._model.eval()
        with torch.no_grad():
            outputs = self._model(img_tensor).squeeze(0)
        k = 1
        if self._args['topk']:
            k = self._args['topk']
        Outputs = NamedTuple('Outputs', [('predict', str)])
        label_l = []
        prob_l = []
        for idx in torch.topk(outputs, k).indices.tolist():
            prob_i = torch.softmax(outputs, -1)[idx].item()
            prob_l.append(prob_i)
            label_i = self._args['labels_map'][idx]
            label_l.append(label_i)
            print(f'[{idx}] {label_i:<75} ({prob_i * 100:.2f}%)')
        return Outputs(label_l[0])



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import threading
import time
from typing import Optional

from towhee.dataframe import DataFrame


class DfWriter(threading.Thread):
    """Put data to dataframe
    """

    def __init__(self, df: DataFrame, count: Optional[int] = None, data=()):
        super().__init__()
        self._df = df
        self._need_stop = False
        self._count = count
        self._set_sealed = False
        self._data = data

    def run(self):
        while not self._need_stop:
            if self._count is not None and self._count <= 0:
                break
            self._df.put(self._data)
            self._count -= 1
            time.sleep(0.02)

        if self._set_sealed:
            self._df.seal()

    def set_sealed_when_stop(self):
        self._set_sealed = True

    def stop(self):
        self.need_stop = True


class MultiThreadRunner:
    """
    Util for multitread test
    """

    def __init__(self, target, args, thread_num):
        self.threads = [threading.Thread(target=target, args=args)
                        for i in range(thread_num)]

    def start(self):
        for t in self.threads:
            t.start()

    def join(self):
        for t in self.threads:
            t.join()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from pathlib import Path

from PIL import Image

from towhee.tests.mock_operators import PYTORCH_CNN_OPERATOR_PATH, load_local_operator
from towhee.tests.mock_operators.pytorch_transform_operator.pytorch_transform_operator import PytorchTransformOperator

cache_path = Path(__file__).parent.parent.resolve()
test_image = cache_path.joinpath('dataset/kaggle_dataset_small/train/001cdf01b096e06d78e9e5112d419397.jpg')


class TestOperator(unittest.TestCase):
    """
    Simple operator test
    """
    def setUp(self) -> None:
        img_pil = Image.open(test_image)
        op = PytorchTransformOperator(256)
        self.img_tensor = op(img_pil).img_transformed
        self.model = 'resnet50'

    def test_func_operator(self):
        pytorch_cnn_operator = load_local_operator(
            'pytorch_cnn_operator', PYTORCH_CNN_OPERATOR_PATH)
        op = pytorch_cnn_operator.PytorchCnnOperator(self.model)
        self.assertEqual((1, 1000), op(self.img_tensor)[0].shape)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from queue import Queue

from towhee.dataframe import DataFrame, Variable
from towhee.engine._operator_io import MapDataFrameReader

from towhee.tests.test_util.dataframe_test_util import DfWriter, MultiThreadRunner


class TestOperatorIO(unittest.TestCase):
    """
    op_ctx IO test
    """

    def test_map_reader(self):
        df = DataFrame('test')
        data = (Variable('int', 1), Variable(
            'str', 'test'), Variable('float', 0.1))
        t = DfWriter(df, 10, data=data)
        t.start()
        t.join()

        map_reader = MapDataFrameReader(df, {'v1': 0, 'v2': 2})
        self.assertEqual(map_reader.size, 10)
        count = 0
        while True:
            item = map_reader.read()
            if not item:
                break
            self.assertEqual(item, {'v1': 1, 'v2': 0.1})
            count += 1
        self.assertEqual(count, 10)
        df.seal()
        self.assertEqual(map_reader.read(), None)

    def test_map_reader_multithread(self):
        df = DataFrame('test')
        data = (Variable('int', 1), Variable(
            'str', 'test'), Variable('float', 0.1))
        data_size = 100
        t = DfWriter(df, data_size, data=data)
        t.set_sealed_when_stop()
        t.start()
        map_reader = MapDataFrameReader(df, {'v1': 0, 'v2': 2})

        q = Queue()

        def read(map_reader: MapDataFrameReader, q: Queue):
            while True:
                item = map_reader.read()
                if item:
                    q.put(item)
                    continue
                elif item is None:
                    break
                else:
                    pass

        runner = MultiThreadRunner(
            target=read, args=(map_reader, q), thread_num=10)

        runner.start()
        runner.join()

        count = 0
        while not q.empty():
            self.assertEqual(q.get(), {'v1': 1, 'v2': 0.1})
            count += 1
        self.assertEqual(count, data_size)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict
import threading

from towhee.dag.operator_repr import OperatorRepr

from towhee.engine.task import Task
# from towhee.engine.graph_context import GraphContext
from towhee.dataframe import DataFrame
from towhee.engine._operator_io import create_reader, create_writer


class OperatorContext:
    """
    The OperatorContext manages an operator's input data and output data at runtime,
    as well as the operators' dependency within a GraphContext.
    The abstraction of OperatorContext hides the complexity of Dataframe management,
    input iteration, and data dependency between Operators. It offers a Task-based
    scheduling context.

    Args:
        op_repr: (OperatorRepr)
            The operator representation
        dataframes: (`dict` of `DataFrame`)
            All the `DataFrames` in `GraphContext`
        is_schedulable: (`bool`)
            Whether the `OperatorContext` is schedulable.
            There are special `OperatorContext`s that are not schedulable, such as
            `_start_op`, `_end_op`.

    """

    def __init__(
        self,
        op_repr: OperatorRepr,
        dataframes: Dict[str, DataFrame],
        is_schedulable: bool = True
    ):
        self._repr = op_repr
        self._is_schedulable = is_schedulable

        # todo: GuoRentong, issue #114
        inputs = list({dataframes[input['df']] for input in op_repr.inputs})
        input_iter_type = op_repr.iter_info['type']
        inputs_index = dict((item['name'], item['col'])
                            for item in op_repr.inputs)
        self.inputs = inputs
        self._reader = create_reader(inputs, input_iter_type, inputs_index)

        outputs = list({dataframes[output['df']]
                       for output in op_repr.outputs})
        self._writer = create_writer(outputs)
        self.outputs = outputs

        self._finished = False
        self._has_tasks = True
        self._taskid = 0
        self._finished_task_count = 0
        self._lock = threading.Lock()

        self.on_start_handlers = []
        self.on_finish_handlers = []

        self.on_task_ready_handlers = []
        self.on_task_start_handlers = []
        self.on_task_finish_handlers = [self._write_outputs]

    @property
    def name(self):
        return self._repr.name

    @property
    def is_schedulable(self) -> bool:
        return self._is_schedulable

    def pop_ready_tasks(self, n_tasks: int = 1) -> List:
        """
        Pop n ready Tasks if any. The number of returned Tasks may be less than n
        if there are not enough Tasks.

        Return: a list of ready Tasks.
        """
        ready_tasks = []
        task_num = n_tasks
        while task_num > 0:
            op_input_params = self._reader.read()
            if op_input_params:
                task = self._create_new_task(op_input_params)
                ready_tasks.append(task)
                task_num -= 1
                continue

            if op_input_params is None:
                self._has_tasks = False

            if not self._has_tasks and self._taskid == self._finished_task_count:
                self._writer.close()
                self._finished = True
            break
        return ready_tasks

    @property
    def is_finished(self) -> bool:
        # todo: GuoRentong. see issue #124
        return self._finished

    @property
    def has_tasks(self) -> bool:
        # todo: GuoRentong. see issue #124
        return self._has_tasks

    @property
    def num_ready_tasks(self) -> int:
        """
        Get the number of ready Tasks.
        """
        return self._reader.size

    @property
    def num_finished_tasks(self) -> int:
        """
        Get the number of finished tasks.
        """
        raise NotImplementedError
        # consider the thread-safe read write. This OperatorContext should be
        # self._finished_tasks' only monifier.

    def _write_outputs(self, task: Task):
        with self._lock:
            self._finished_task_count += 1
            self._writer.write(task.outputs)

    def _next_task_inputs(self):
        """
        Manage the preparation works for an operator's inputs
        Returns one task's inputs on each call.

        Return: a list of inputs, list element can be scalar or Array.
        """
        raise NotImplementedError

    def _create_new_task(self, inputs: Dict[str, any]):
        with self._lock:
            t = Task(self.name, self._repr.function,
                     self._repr.init_args, inputs, self._taskid)
            self._taskid += 1
            t.add_finish_handler(self.on_task_finish_handlers)
            return t



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from towhee.array.array import Array


def full(shape: int, fill_value):
    """Return a new `Array` of given shape filled with `fill_value`

    Args:
        shape: (`int`)
            Shape of the new `Array`
        fill_value: (`numpy.dtype` or `str`)
            Fill value
    """
    return Array(data=np.full(shape=shape, fill_value=fill_value))



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import NamedTuple

from towhee.operator import Operator, SharedType


class SubOperator(Operator):
    """Subtract operator (used for unit test purposes only).
    """

    def __init__(self):
        super().__init__()
        self.key = ()

    def __call__(self, a: int, b: int) -> NamedTuple('Outputs', [('diff', int)]):
        Outputs = NamedTuple('Outputs', [('diff', int)])
        return Outputs(a - b)

    @property
    def shared_type(self):
        return SharedType.Shareable



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from functools import wraps
from variable_repr import VariableReprSet


def op_action(func):
    """
    Mark of the implementation of an Operator's callable part.

    Examples:
        Only interface Operator will use the @op_action mark. In most cases, method
        decorated by @op_action will be called by Operator's __call__. During an
        Operator's creation, OpMetaclass will check that method with @op_action in the
        interface Operator is overwritten by the subclass with identical input/output
        declarations.

        Example of an interface Operator

        class Img2VecOp(Operator):
            @op_action
            @abc.abstractmethod
            def forward(self, img:Image) -> FloatVector:
                pass

            @create_op_in_pipeline
            def __call__(self, img:Image):
                return forward(img)

        Example of a inherited Operator

        class VGGAnimalImg2VecOp(Img2VecOp):
            def forward(self, img:Image) -> FloatVector:
                # do the real forward works ...

        Note that Img2VecOp.forward is decorated by @op_action, and VGGAnimalImg2VecOp
        implements this function with identical input types and output types.

        See OpMetaclass._op_callable_check for more explainations.
    """
    raise NotImplementedError


def create_op_in_pipeline(func):
    """
    This is a decorator of the operator's __call__ method. It will put the operator
    into the pipeline's context -- creating a node related to the operator, and solving
    its dependency.

    Example:
        class MyOp1(Operator):
            @create_op_in_pipeline
            def __call__(self, x:int) -> NamedTuple('Outputs', [('y1', float), ('y2', str)]):
                return forward(img)

        class MyOp2(Operator):
            @create_op_in_pipeline
            def __call__(self, x:float) -> float:
                return forward(img)

        @create_pipeline
        def my_pipeline(x:int):
            op1 = MyOp1()
            op2 = MyOp2()
            y = op1(x)
            z = op2(y.y1)
            return z

    In this example, we have two operators in a pipeline, where op2 depends on the
    results of op1. Behind the scene, Towhee's compiler will create a DAG during
    the execution of my_pipeline. When the program executes the line y = op1(x),
    The decorator @create_op_in_pipeline will be called before MyOp1's __call__
    method. It will put op1 into the pipeline's context, link the pipeline's input x
    to op1's input. When the program reaches the line z = op2(y.y1), the compiler will
    add op2 into the pipeline's context, and settle the dependency between op1 and op2
    base on the fact that op2 takes op1's output as its input.
    """
    @wraps(func)
    def _create_op_in_pipeline(*args, **kwargs) -> VariableReprSet:
        """
        Solving the operator's dependency and return a VariableReprSet as the operator's
        outputs.
        """
        #raise NotImplementedError
        return func(*args, **kwargs)

    return _create_op_in_pipeline



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from pathlib import Path

from towhee.operator import Operator
from towhee.engine.operator_pool import OperatorPool
from towhee.engine.task import Task


class TestOperatorPool(unittest.TestCase):
    """Basic test case for `OperatorPool`.
    """

    def setUp(self):
        cache_path = Path(__file__).parent.parent.resolve()
        self._op_pool = OperatorPool(cache_path=cache_path)

    def test_init(self):
        # The list of available ops should be empty upon initialization.
        self.assertFalse(self._op_pool.available_ops)

    def test_acquire_release(self):

        hub_op_id = 'mock_operators/add_operator'
        task = Task('test', hub_op_id, {'factor': 0}, (1), 0)

        # Acquire the operator.
        op = self._op_pool.acquire_op(task)

        # Perform some simple operations.
        self.assertTrue(isinstance(op, Operator))
        self.assertEqual(op(1).sum, 1)

        # Release and re-acquire the operator.
        self._op_pool.release_op(op)
        op = self._op_pool.acquire_op(task)

        # Perform more operations.
        self.assertEqual(op(-1).sum, -1)
        self.assertEqual(op(100).sum, 100)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path

_DEFAULT_LOCAL_CACHE_ROOT = Path.home() / '.towhee'
LOCAL_PIPELINE_CACHE = _DEFAULT_LOCAL_CACHE_ROOT / 'pipelines'
LOCAL_OPERATOR_CACHE = _DEFAULT_LOCAL_CACHE_ROOT / 'operators'

MOCK_PIPELINES = str(Path(__file__).parent.parent /
                     'tests/test_util/resnet50_embedding.yaml')

MOCK_OPS = str(Path(__file__).parent.parent / 'tests/mock_operators')



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import dataclasses
import json

from dataclasses import dataclass
from typing import Dict, List, Optional, Union
from tqdm.auto import tqdm
from towhee.cnn_trainer.training_args import TrainingArguments
from towhee.cnn_trainer.utils import logging

logger = logging.get_logger(__name__)


@dataclass
class TrainerState:
    """
    A class containing the :class:`~PyTorchCNNTrainer` inner state that will be saved along the model and optimizer
    when checkpointing and passed to the :class:`~towhee.TrainerCallback`.
    Args:
        epoch (:obj:`float`, `optional`):
            Only set during training, will represent the epoch the training is at (the decimal part being the
            percentage of the current epoch completed).
        global_step (:obj:`int`, `optional`, defaults to 0):
            During training, represents the number of update steps completed.
        max_steps (:obj:`int`, `optional`, defaults to 0):
            The number of update steps to do during the current training.
        total_flos (:obj:`float`, `optional`, defaults to 0):
            The total number of floating operations done by the model since the beginning of training (stored as floats
            to avoid overflow).
        log_history (:obj:`List[Dict[str, float]]`, `optional`):
            The list of logs done since the beginning of training.
        best_metric (:obj:`float`, `optional`):
            When tracking the best model, the value of the best metric encountered so far.
        best_model_checkpoint (:obj:`str`, `optional`):
            When tracking the best model, the value of the name of the checkpoint for the best model encountered so
            far.
        is_local_process_zero (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Whether or not this process is the local (e.g., on one machine if training in a distributed fashion on
            several machines) main process.
        is_world_process_zero (:obj:`bool`, `optional`, defaults to :obj:`True`):
            Whether or not this process is the global main process (when training in a distributed fashion on several
            machines, this is only going to be :obj:`True` for one process).
        is_hyper_param_search (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether we are in the process of a hyper parameter search using Trainer.hyperparameter_search. This will
            impact the way data will be logged in TensorBoard.
    """

    epoch: Optional[float] = None
    global_step: int = 0
    max_steps: int = 0
    num_train_epochs: int = 0
    total_flos: float = 0
    log_history: List[Dict[str, float]] = None
    best_metric: Optional[float] = None
    best_model_checkpoint: Optional[str] = None
    is_local_process_zero: bool = True
    is_world_process_zero: bool = True
    is_hyper_param_search: bool = False
    trial_name: str = None
    trial_params: Dict[str, Union[str, float, int, bool]] = None

    def __post_init__(self):
        if self.log_history is None:
            self.log_history = []

    def save_to_json(self, json_path: str):
        """Save the content of this instance in JSON format inside :obj:`json_path`."""
        json_string = json.dumps(dataclasses.asdict(self), indent=2, sort_keys=True) + "\n"
        with open(json_path, "w", encoding="utf-8") as f:
            f.write(json_string)

    @classmethod
    def load_from_json(cls, json_path: str):
        """Create an instance from the content of :obj:`json_path`."""
        with open(json_path, "r", encoding="utf-8") as f:
            text = f.read()
        return cls(**json.loads(text))


@dataclass
class TrainerControl:
    """
    A class that handles the :class:`~PyTorchCNNTrainer` control flow. This class is used by the
    :class:`~towhee.TrainerCallback` to activate some switches in the training loop.
    Args:
        should_training_stop (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether or not the training should be interrupted.
            If :obj:`True`, this variable will not be set back to :obj:`False`. The training will just stop.
        should_epoch_stop (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether or not the current epoch should be interrupted.
            If :obj:`True`, this variable will be set back to :obj:`False` at the beginning of the next epoch.
        should_save (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether or not the model should be saved at this step.
            If :obj:`True`, this variable will be set back to :obj:`False` at the beginning of the next step.
        should_evaluate (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether or not the model should be evaluated at this step.
            If :obj:`True`, this variable will be set back to :obj:`False` at the beginning of the next step.
        should_log (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether or not the logs should be reported at this step.
            If :obj:`True`, this variable will be set back to :obj:`False` at the beginning of the next step.
    """

    should_training_stop: bool = False
    should_epoch_stop: bool = False
    should_save: bool = False
    should_evaluate: bool = False
    should_log: bool = False

    def _new_training(self):
        """Internal method that resets the variable for a new training."""
        self.should_training_stop = False

    def _new_epoch(self):
        """Internal method that resets the variable for a new epoch."""
        self.should_epoch_stop = False

    def _new_step(self):
        """Internal method that resets the variable for a new step."""
        self.should_save = False
        self.should_evaluate = False
        self.should_log = False


class TrainerCallback:
    """
    A class for objects that will inspect the state of the training loop at some events and take some decisions. At
    each of those events the following arguments are available:
    Args:
        args (:class:`~towhee.TrainingArguments`):
            The training arguments used to instantiate the :class:`~towhee.Trainer`.
        state (:class:`~towhee.TrainerState`):
            The current state of the :class:`~towhee.Trainer`.
        control (:class:`~towhee.TrainerControl`):
            The object that is returned to the :class:`~towhee.Trainer` and can be used to make some decisions.
        model (:obj:`torch.nn.Module`):
            The model being trained.
        optimizer (:obj:`torch.optim.Optimizer`):
            The optimizer used for the training steps.
        lr_scheduler (:obj:`torch.optim.lr_scheduler.LambdaLR`):
            The scheduler used for setting the learning rate.
        train_dataloader (:obj:`torch.utils.data.dataloader.DataLoader`, `optional`):
            The current dataloader used for training.
        eval_dataloader (:obj:`torch.utils.data.dataloader.DataLoader`, `optional`):
            The current dataloader used for training.
        logs  (:obj:`Dict[str, float]`):
            The values to log.
            Those are only accessible in the event :obj:`on_log`.
    The :obj:`control` object is the only one that can be changed by the callback, in which case the event that changes
    it should return the modified version.
    The argument :obj:`args`, :obj:`state` and :obj:`control` are positionals for all events, all the others are
    grouped in :obj:`kwargs`. You can unpack the ones you need in the signature of the event using them. As an example,
    see the code of the simple :class:`~towhee.PrinterCallback`.
    """

    def on_init_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of the initialization of the :class:`~towhee.Trainer`.
        """
        pass

    def on_train_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the beginning of training.
        """
        pass

    def on_train_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of training.
        """
        pass

    def on_epoch_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the beginning of an epoch.
        """
        pass

    def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of an epoch.
        """
        pass

    def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the beginning of a training step. If using gradient accumulation, one training step might take
        several inputs.
        """
        pass

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called at the end of a training step. If using gradient accumulation, one training step might take
        several inputs.
        """
        pass

    def on_evaluate(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after an evaluation phase.
        """
        pass

    def on_save(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after a checkpoint save.
        """
        pass

    def on_log(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after logging the last logs.
        """
        pass

    def on_prediction_step(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        """
        Event called after a prediction step.
        """
        pass


class CallbackHandler(TrainerCallback):
    """Internal class that just calls the list of callbacks in order."""

    def __init__(self, callbacks, model, optimizer, lr_scheduler):
        self.callbacks = []
        for cb in callbacks:
            self.add_callback(cb)
        self.model = model
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.train_dataloader = None
        self.eval_dataloader = None

        if not any(isinstance(cb, DefaultFlowCallback) for cb in self.callbacks):
            logger.warning(
                "Unexpected error."
            )

    def add_callback(self, callback):
        cb = callback() if isinstance(callback, type) else callback
        cb_class = callback if isinstance(callback, type) else callback.__class__
        if cb_class in [c.__class__ for c in self.callbacks]:
            logger.warning(
                "add_callback error."
            )
        self.callbacks.append(cb)

    def pop_callback(self, callback):
        if isinstance(callback, type):
            for cb in self.callbacks:
                if isinstance(cb, callback):
                    self.callbacks.remove(cb)
                    return cb
        else:
            for cb in self.callbacks:
                if cb == callback:
                    self.callbacks.remove(cb)
                    return cb

    def remove_callback(self, callback):
        if isinstance(callback, type):
            for cb in self.callbacks:
                if isinstance(cb, callback):
                    self.callbacks.remove(cb)
                    return
        else:
            self.callbacks.remove(callback)

    @property
    def callback_list(self):
        return "\n".join(cb.__class__.__name__ for cb in self.callbacks)

    def on_init_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        return self.call_event("on_init_end", args, state, control)

    def on_train_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        control.should_training_stop = False
        return self.call_event("on_train_begin", args, state, control)

    def on_train_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        return self.call_event("on_train_end", args, state, control)

    def on_epoch_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        control.should_epoch_stop = False
        return self.call_event("on_epoch_begin", args, state, control)

    def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        return self.call_event("on_epoch_end", args, state, control)

    def on_step_begin(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        control.should_log = False
        control.should_evaluate = False
        control.should_save = False
        return self.call_event("on_step_begin", args, state, control)

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        return self.call_event("on_step_end", args, state, control)

    def on_evaluate(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, metrics):
        control.should_evaluate = False
        return self.call_event("on_evaluate", args, state, control, metrics=metrics)

    def on_save(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        control.should_save = False
        return self.call_event("on_save", args, state, control)

    def on_log(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, logs):
        control.should_log = False
        return self.call_event("on_log", args, state, control, logs=logs)

    def on_prediction_step(self, args: TrainingArguments, state: TrainerState, control: TrainerControl):
        return self.call_event("on_prediction_step", args, state, control)

    def call_event(self, event, args, state, control, **kwargs):
        for callback in self.callbacks:
            result = getattr(callback, event)(
                args,
                state,
                control,
                model=self.model,
                optimizer=self.optimizer,
                lr_scheduler=self.lr_scheduler,
                train_dataloader=self.train_dataloader,
                eval_dataloader=self.eval_dataloader,
                **kwargs,
            )
            # A Callback can skip the return of `control` if it doesn't change it.
            if result is not None:
                control = result
        return control


class DefaultFlowCallback(TrainerCallback):
    """
    A :class:`~towhee.TrainerCallback` that handles the default flow of the training loop for logs, evaluation
    and checkpoints.
    """

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        # Log
        control.should_log = True

        # Save
        control.should_save = True

        return control

    def on_epoch_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        # Log
        control.should_log = True

        # Evaluate
        control.should_save = True

        return control


class ProgressCallback(TrainerCallback):
    """
    A :class:`~towhee.TrainerCallback` that displays the progress of training or evaluation.
    """

    def __init__(self):
        self.training_bar = None
        self.prediction_bar = None

    def on_train_begin(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            self.training_bar = tqdm(total=state.max_steps)
        self.current_step = 0

    def on_step_end(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            self.training_bar.update(state.global_step - self.current_step)
            self.current_step = state.global_step

    def on_prediction_step(self, args, state, control, eval_dataloader=None, **kwargs):
        if state.is_local_process_zero and isinstance(eval_dataloader.dataset, collections.abc.Sized):
            if self.prediction_bar is None:
                self.prediction_bar = tqdm(total=len(eval_dataloader), leave=self.training_bar is None)
            self.prediction_bar.update(1)

    def on_evaluate(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            if self.prediction_bar is not None:
                self.prediction_bar.close()
            self.prediction_bar = None

    def on_log(self, args, state, control, logs=None, **kwargs):
        if state.is_local_process_zero and self.training_bar is not None:
            _ = logs.pop("total_flos", None)
            self.training_bar.write(str(logs))

    def on_train_end(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            self.training_bar.close()
            self.training_bar = None


class PrinterCallback(TrainerCallback):
    """
    A bare :class:`~towhee.TrainerCallback` that just prints the logs.
    """

    def on_log(self, args, state, control, logs=None, **kwargs):
        _ = logs.pop("total_flos", None)
        if state.is_local_process_zero:
            print(logs)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT_ WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from towhee.engine.operator_context import OperatorContext
from towhee.dataframe import DataFrame, Variable
from towhee.dag import OperatorRepr

from towhee.tests.test_util.dataframe_test_util import DfWriter


class TestOperatorContext(unittest.TestCase):
    """UT for `OperatorContext`
    """

    def test_op_context(self):

        df_in = DataFrame('op_test_in')
        df_out = DataFrame('op_test_out')
        dfs = {'op_test_in': df_in, 'op_test_out': df_out}

        op_repr = OperatorRepr(
            name='mock_op',
            function='mock_op',
            init_args=None,
            inputs=[
                {'name': 'k1', 'df': 'op_test_in', 'col': 0},
                {'name': 'k2', 'df': 'op_test_in', 'col': 1}
            ],
            outputs=[{'df': 'op_test_out'}],
            iter_info={'type': 'map'}
        )

        op = OperatorContext(op_repr, dfs)
        self.assertEqual(len(op.pop_ready_tasks()), 0)

        data = (Variable('int', 1), Variable(
            'str', 'test'), Variable('float', 0.1))
        data_size = 20
        t = DfWriter(df_in, data_size, data=data)
        t.set_sealed_when_stop()
        t.start()
        t.join()

        count = 0
        while True:
            tasks = op.pop_ready_tasks(2)
            if not op.has_tasks:
                break
            self.assertEqual(len(tasks), 2)
            count += len(tasks)
            self.assertEqual(op.num_ready_tasks, data_size - count)

        self.assertEqual(count, data_size)

    def test_op_context_multithread(self):

        df_in = DataFrame('op_test_in')
        df_out = DataFrame('op_test_out')
        dfs = {'op_test_in': df_in, 'op_test_out': df_out}

        op_repr = OperatorRepr(
            name='mock_op',
            function='mock_op',
            init_args=None,
            inputs=[
                {'name': 'k1', 'df': 'op_test_in', 'col': 0},
                {'name': 'k2', 'df': 'op_test_in', 'col': 1}
            ],
            outputs=[{'df': 'op_test_out'}],
            iter_info={'type': 'map'}
        )

        op = OperatorContext(op_repr, dfs)
        self.assertEqual(len(op.pop_ready_tasks()), 0)

        data = (Variable('int', 1), Variable(
            'str', 'test'), Variable('float', 0.1))
        data_size = 20
        t = DfWriter(df_in, data_size, data=data)
        t.set_sealed_when_stop()
        t.start()

        count = 0
        while True:
            tasks = op.pop_ready_tasks(5)
            if not op.has_tasks:
                break
            self.assertLessEqual(len(tasks), 2)
            count += len(tasks)

        self.assertEqual(count, data_size)
